"""Tests for Trained Model."""

from unittest.mock import patch

import numpy as np
from sotai import (
    DatasetSplit,
    FeatureType,
    LinearConfig,
    LossType,
    Metric,
    NumericalFeatureConfig,
    PipelineConfig,
    TargetType,
    TrainedModel,
    TrainedModelMetadata,
    TrainingConfig,
    TrainingResults,
)
from sotai.enums import APIStatus
from sotai.models import CalibratedLinear

from .utils import construct_trained_model


def test_trained_classification_model_predict(fixture_data, fixture_feature_configs):
    """Tests the predict function on a trained model."""
    trained_model = construct_trained_model(
        TargetType.CLASSIFICATION, fixture_data, fixture_feature_configs
    )
    predictions, probabilities = trained_model.predict(fixture_data)
    assert isinstance(predictions, np.ndarray)
    assert len(predictions) == len(fixture_data)
    assert isinstance(probabilities, np.ndarray)
    assert len(probabilities) == len(fixture_data)


def test_trained_regression_model_predict(fixture_data, fixture_feature_configs):
    """Tests the predict function on a trained model."""
    trained_model = construct_trained_model(
        TargetType.REGRESSION, fixture_data, fixture_feature_configs
    )
    predictions, _ = trained_model.predict(fixture_data)
    assert isinstance(predictions, np.ndarray)
    assert len(predictions) == len(fixture_data)


def test_trained_model_save_load(
    fixture_data,
    fixture_feature_configs,
    tmp_path,
):
    """Tests that a `TrainedModel` can be successfully saved and then loaded."""
    trained_model = construct_trained_model(
        TargetType.CLASSIFICATION, fixture_data, fixture_feature_configs
    )
    trained_model.save(tmp_path)
    loaded_trained_model = TrainedModel.load(tmp_path)
    assert isinstance(loaded_trained_model, TrainedModel)
    assert loaded_trained_model.dict(
        exclude={"model", "saved_filepath"}
    ) == trained_model.dict(exclude={"model", "saved_filepath"})
    assert isinstance(loaded_trained_model.model, CalibratedLinear)


@patch("sotai.trained_model.download_trained_model")
@patch(
    "sotai.trained_model.get_trained_model_metadata",
    return_value=(
        APIStatus.SUCCESS,
        TrainedModelMetadata(
            **{
                "id": 0,
                "dataset_id": 0,
                "model_config": LinearConfig(),
                "training_config": TrainingConfig(
                    loss_type=LossType.MSE,
                    epochs=1,
                    batch_size=1,
                    learning_rate=1,
                ),
                "training_results": TrainingResults(
                    training_time=1,
                    train_loss_by_epoch=[1],
                    train_primary_metric_by_epoch=[1],
                    val_loss_by_epoch=[1],
                    val_primary_metric_by_epoch=[1],
                    evaluation_time=1,
                    test_loss=1,
                    test_primary_metric=1,
                    feature_analyses={},
                    linear_coefficients={},
                ),
                "uuid": "test_uuid",
                "pipeline_config": PipelineConfig(
                    id=0,
                    target="target",
                    target_type=TargetType.CLASSIFICATION,
                    primary_metric=Metric.MSE,
                    feature_configs={
                        "age": NumericalFeatureConfig(
                            name="age",
                            type=FeatureType.NUMERICAL,
                        ),
                    },
                    shuffle_data=False,
                    drop_empty_percentage=70,
                    dataset_split=DatasetSplit(train=80, val=10, test=10),
                ),
            }
        ),
    ),
)
def test_load_from_hosted(
    mock_get_trained_model,
    mock_download_trained_model,
    tmp_path,
    fixture_data,
    fixture_feature_configs,
):
    """Test loading a trained model from hosted."""
    trained_model = construct_trained_model(
        TargetType.CLASSIFICATION, fixture_data, fixture_feature_configs
    )
    trained_model.save(tmp_path)
    mock_download_trained_model.return_value = f"{tmp_path}/trained_ptcm_model.pt"
    trained_model = TrainedModel.from_hosted("test_uuid")
    mock_get_trained_model.assert_called_once()
    mock_download_trained_model.assert_called_once()
    assert isinstance(trained_model, TrainedModel)
    assert trained_model.training_config.epochs == 1
    assert trained_model.training_config.batch_size == 1
    assert trained_model.training_config.learning_rate == 1
    assert trained_model.training_config.loss_type == LossType.MSE
    assert trained_model.training_results.training_time == 1
    assert trained_model.training_results.train_loss_by_epoch == [1]
    assert trained_model.training_results.train_primary_metric_by_epoch == [1]
    assert trained_model.training_results.val_loss_by_epoch == [1]
    assert trained_model.model_config.output_calibration is False



"""This module contains the API functions for interacting with the SOTAI API.""" ""
import logging
import os
import tarfile
import urllib
from typing import Any, Dict, List, Optional, Tuple, Type, Union
import importlib.metadata
import pandas as pd
import requests

from .constants import SOTAI_API_ENDPOINT, SOTAI_API_TIMEOUT, SOTAI_BASE_URL
from .enums import APIStatus, InferenceConfigStatus
from .types import (
    Dataset,
    PreparedData,
    CategoricalFeatureConfig,
    DatasetSplit,
    FeatureType,
    HypertuneConfig,
    LinearConfig,
    NumericalFeatureConfig,
    PipelineConfig,
    TrainedModelMetadata,
    TrainingConfig,
    TrainingResults,
    _BaseModelConfig,
)


def set_api_key(api_key: str):
    """Set the SOTAI API key in the environment variables.

    Args:
        api_key: The API key to set.
    """
    os.environ["SOTAI_API_KEY"] = api_key


def get_api_key() -> str:
    """Returns the SOTAI API key from the environment variables."""
    return os.environ["SOTAI_API_KEY"]


def get_auth_headers() -> Dict[str, str]:
    """Returns the authentication headers for a pipeline."""
    return {
        "sotai-api-key": get_api_key(),
        "sotai-sdk-version": importlib.metadata.version("sotai"),
    }


def extract_response(
    api_call: str, response: requests.Response
) -> Tuple[APIStatus, Optional[Any]]:
    """Extract the response from a requests response.

    Args:
        api_call: The name of the API call.
        response: The requests response.

    Returns:
        A tuple containing the status of the API call and JSON response encoded content.
        If unsuccessful, the response will be `None`.
    """
    if response.status_code != 200:
        if response.headers.get("content-type") == "application/json":
            logging.error("API call %s failed. %s", api_call, response.json()["detail"])
        else:
            logging.error("API call %s failed.", api_call)
        return APIStatus.ERROR, None
    return APIStatus.SUCCESS, response.json()


def post_pipeline(pipeline) -> Tuple[APIStatus, Optional[str]]:
    """Create a new pipeline on the SOTAI API.

    Args:
        pipeline: The pipeline to create.

    Returns:
        A tuple containing the status of the API call and the UUID of the created
        pipeline. If unsuccessful, the UUID will be `None`.
    """
    response = requests.post(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipelines",
        json={
            "name": pipeline.name,
            "target": pipeline.target,
            "target_column_type": pipeline.target_type,
            "primary_metric": pipeline.primary_metric,
        },
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    api_status, response_json = extract_response("post_pipeline", response)
    if api_status == APIStatus.ERROR:
        return api_status, None

    return api_status, response_json["uuid"]


def post_pipeline_config(
    pipeline_uuid: str, pipeline_config: PipelineConfig
) -> Tuple[APIStatus, Optional[str]]:
    """Create a new pipeline config on the SOTAI API.

    Args:
        pipeline_uuid: The pipeline uuid to create the pipeline config for.
        pipeline_config : The pipeline config to create.

    Returns:
        A tuple containing the status of the API call and the UUID of the created
        pipeline. If unsuccessful, the UUID will be `None`.
    """
    response = requests.post(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipelines/{pipeline_uuid}/pipeline-configs",
        json={
            "shuffle_data": pipeline_config.shuffle_data,
            "drop_empty_percentage": pipeline_config.drop_empty_percentage,
            "train_percentage": pipeline_config.dataset_split.train,
            "validation_percentage": pipeline_config.dataset_split.val,
            "test_percentage": pipeline_config.dataset_split.test,
            "pipeline_config_sdk_id": pipeline_config.id,
        },
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    api_status, response_json = extract_response("post_pipeline_config", response)
    if api_status == APIStatus.ERROR:
        return api_status, None

    return APIStatus.SUCCESS, response_json["uuid"]


def post_pipeline_feature_configs(
    pipeline_config_uuid: str,
    feature_configs: Dict[str, Union[CategoricalFeatureConfig, NumericalFeatureConfig]],
) -> APIStatus:
    """Create a new pipeline feature configs on the SOTAI API.

    Args:
        pipeline_config_uuid: The pipeline config uuid to create the pipeline
            feature configs for.
        feature_configs: The feature configs to create.

    Returns:
        The status of the API call.
    """
    sotai_feature_configs = []

    for feature_config in feature_configs.values():
        sotai_feature_config = {
            "feature_name": feature_config.name,
            "feature_type": feature_config.type,
        }
        if feature_config.type == FeatureType.CATEGORICAL:
            if isinstance(feature_config.categories[0], int):
                sotai_feature_config["categories_int"] = feature_config.categories
            else:
                sotai_feature_config["categories_str"] = feature_config.categories
        else:
            sotai_feature_config["num_keypoints"] = feature_config.num_keypoints
            sotai_feature_config[
                "input_keypoints_init"
            ] = feature_config.input_keypoints_init
            sotai_feature_config[
                "input_keypoints_type"
            ] = feature_config.input_keypoints_type
            sotai_feature_config["monotonicity"] = feature_config.monotonicity

        sotai_feature_configs.append(sotai_feature_config)

    response = requests.post(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipeline-configs/"
        f"{pipeline_config_uuid}/feature-configs",
        json=sotai_feature_configs,
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    api_status, _ = extract_response("post_pipeline_feature_configs", response)

    return api_status


def post_trained_model_analysis(
    pipeline_config_uuid: str, trained_model: TrainedModelMetadata
) -> Tuple[APIStatus, Optional[Dict[str, str]]]:
    """Create a new trained model analysis on the SOTAI API.

    Args:
        pipeline_config_uuid: The pipeline config uuid to create the trained model
            analysis for.
        trained_model: The trained model to create.

    Returns:
        A tuple containing the status of the API call and a dict containing the UUIDs
        of the resources created as well as a link that can be used to view the trained
        model analysis. If unsuccessful, the UUID will be `None`.

        Keys:
            - `trained_model_metadata_uuid`: The UUID of the trained model.
            - `model_config_uuid`: The UUID of the model configuration.
            - `pipeline_config_uuid`: The UUID of the pipeline configuration.
            - `analysis_url`: The URL of the trained model analysis.
    """
    training_results = trained_model.training_results
    train_primary_metrics = training_results.train_primary_metric_by_epoch
    val_primary_metrics = training_results.val_primary_metric_by_epoch
    overall_model_results_dict = {
        "epochs": trained_model.training_config.epochs,
        "batch_size": trained_model.training_config.batch_size,
        "learning_rate": trained_model.training_config.learning_rate,
        "runtime_in_seconds": round(training_results.training_time),
        "train_loss_per_epoch": training_results.train_loss_by_epoch,
        "train_primary_metric_per_epoch": train_primary_metrics,
        "validation_loss_per_epoch": training_results.val_loss_by_epoch,
        "validation_primary_metric_per_epoch": val_primary_metrics,
        "test_loss": training_results.test_loss,
        "test_primary_metric": training_results.test_primary_metric,
        "feature_names": [
            feature.feature_name for feature in trained_model.model.features
        ],
        "linear_coefficients": [
            training_results.linear_coefficients[feature.feature_name]
            for feature in trained_model.model.features
        ],
    }
    feature_analyses_list = [
        {
            "feature_name": feature.feature_name,
            "feature_type": feature.feature_type.value,
            "statistic_min": feature.min,
            "statistic_max": feature.max,
            "statistic_mean": feature.mean,
            "statistic_median": feature.median,
            "statistic_std": feature.std,
            "keypoints_outputs": feature.keypoints_outputs,
            "keypoints_inputs_categorical": feature.keypoints_inputs_categorical,
            "keypoints_inputs_numerical": feature.keypoints_inputs_numerical,
        }
        for feature in trained_model.training_results.feature_analyses.values()
    ]
    trained_model_metadata_dict = {
        "epochs": trained_model.training_config.epochs,
        "batch_size": trained_model.training_config.batch_size,
        "learning_rate": trained_model.training_config.learning_rate,
        "train_primary_metric": [training_results.train_primary_metric_by_epoch[-1]],
        "validation_primary_metric": [training_results.val_primary_metric_by_epoch[-1]],
        "test_primary_metric": training_results.test_primary_metric,
        "trained_model_sdk_id": trained_model.id,
    }

    model_config_dict = {
        "model_framework": "pytorch",
        "model_type": "linear",
        "loss_type": trained_model.training_config.loss_type.value,
        "primary_metric": trained_model.pipeline_config.primary_metric.value,
        "target_column_type": trained_model.pipeline_config.target_type.value,
        "target_column": trained_model.pipeline_config.target,
        "model_config_name": f"Model {trained_model.id}",
    }

    response = requests.post(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipeline-configs/{pipeline_config_uuid}/analysis",
        json={
            "trained_model_metadata": trained_model_metadata_dict,
            "overall_model_results": overall_model_results_dict,
            "model_config": model_config_dict,
            "feature_analyses": feature_analyses_list,
        },
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    return extract_response("post_trained_model_analysis", response)


def post_trained_model(trained_model_path: str, trained_model_uuid: str) -> APIStatus:
    """Create a new trained model on the SOTAI API.

    Args:
        trained_model_path: The path to the trained model file to post.
        trained_model_uuid: The UUID of the trained model.

    Returns:
        The status of the API call.
    """
    original_filepath = f"{trained_model_path}/trained_ptcm_model.pt"
    tar_filepath = f"{trained_model_path}/model.tar.gz"
    with tarfile.open(tar_filepath, "w:gz") as tar:
        tar.add(original_filepath, arcname="model.pt")

    with open(tar_filepath, "rb") as data_file:
        response = requests.post(
            f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/models",
            files={"file": data_file},
            data={"trained_model_metadata_uuid": trained_model_uuid},
            headers=get_auth_headers(),
            timeout=SOTAI_API_TIMEOUT,
        )

    api_status, _ = extract_response("post_trained_model", response)
    return api_status


def post_inference(
    data_filepath: str,
    trained_model_uuid: str,
) -> Tuple[APIStatus, Optional[str]]:
    """Create a new inference on the SOTAI API .

    Args:
        data_filepath: The path to the data file to create the inference for.
        trained_model_uuid: The trained model uuid to create the inference for.

    Returns:
        A tuple containing the status of the API call and the UUID of the created
        inference job. If unsuccessful, the UUID will be `None`.
    """
    with open(data_filepath, "rb") as data_file:
        response = requests.post(
            f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/inferences",
            files={"file": data_file},
            data={"trained_model_metadata_uuid": trained_model_uuid},
            headers=get_auth_headers(),
            timeout=SOTAI_API_TIMEOUT,
        )

    api_status, response_json = extract_response("post_trained_model", response)
    if api_status == APIStatus.ERROR:
        return api_status, None

    return api_status, response_json["inference_config_uuid"]


def get_inference_status(
    inference_uuid: str,
) -> Tuple[APIStatus, Optional[InferenceConfigStatus]]:
    """Get an inference from the SOTAI API.

    Args:
        inference_uuid: The UUID of the inference to get.

    Returns:
       A tuple containing the status of the API call and the status of the inference job
       if the API call is successful. If unsuccessful, the UUID will be `None`.
    """
    response = requests.get(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/inferences/{inference_uuid}/status",
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    return extract_response("get_inference_status", response)


def get_inference_results(inference_uuid: str, download_folder: str) -> APIStatus:
    """Get an inference from the SOTAI API.

    Args:
        inference_uuid: The UUID of the inference results to get.

    Returns:
        The status of the API call.
    """
    response = requests.get(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/inferences/{inference_uuid}/download",
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    api_status, response_json = extract_response("get_inference_results", response)
    if api_status == APIStatus.ERROR:
        return api_status

    urllib.request.urlretrieve(
        response_json, f"{download_folder}/inference_results.csv"
    )
    return api_status


def post_dataset(
    train_filepath: str,
    test_filepath: str,
    validation_filepath: str,
    columns: List[str],
    categorical_columns: List[str],
    pipeline_config_uuid: str,
    dataset_id: int,
) -> Tuple[APIStatus, Optional[str]]:
    """Upload a dataset to th the SOTAI API.

    Args:
        data_filepath: The path to the data file to push to the API.
        columns: The columns of the dataset.
        categorical_columns: The categorical columns of the dataset.
        pipeline_uuid: The pipeline uuid for which to upload the dataset.

    Returns:
        A tuple containing the status of the API call and the UUID of the created
        dataset. If unsuccessful, the UUID will be `None`.
    """

    response = requests.post(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/datasets",
        # pylint: disable=consider-using-with
        files=[
            ("files", open(test_filepath, "rb")),
            (
                "files",
                open(train_filepath, "rb"),
            ),
            (
                "files",
                open(validation_filepath, "rb"),
            ),
        ],
        # pylint: enable=consider-using-with
        data={
            "pipeline_config_uuid": pipeline_config_uuid,
            "columns": columns,
            "categorical_columns": categorical_columns,
            "dataset_sdk_id": dataset_id,
        },
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    api_status, response_json = extract_response("post_dataset", response)
    if api_status == APIStatus.ERROR:
        return api_status, None
    return api_status, response_json["uuid"]


def post_hypertune_job(
    hypertune_config: HypertuneConfig,
    pipeline_config: PipelineConfig,
    model_config: Type[_BaseModelConfig],
    dataset_uuid: str,
    next_model_id: int,
):
    """Upload a dataset to th the SOTAI API.

    Args:
        hypertune_config: The hypertune config to create the hypertune job for.
        pipeline_config: The pipeline config to create the hypertune job for.
        dataset_uuid: The dataset uuid to create the hypertune job for.

    Returns:
        A tuple containing the status of the API call and an array of the UUIDs of the
        created trained models. If unsuccessful, the UUIDs will be `None`.
    """

    input_keypoints_type = model_config.output_calibration_input_keypoints_type
    advanced_options = {
        "output_min": model_config.output_min,
        "output_max": model_config.output_max,
        "output_calibration": model_config.output_calibration,
        "output_calibration_num_keypoints": model_config.output_calibration_num_keypoints,
        "output_initialization": model_config.output_initialization,
        "output_calibration_input_keypoints_type": input_keypoints_type,
    }
    if isinstance(model_config, LinearConfig):
        advanced_options["use_bias"] = model_config.use_bias

    response = requests.post(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipeline-config/{pipeline_config.uuid}/hypertune",
        json={
            "dataset_uuid": dataset_uuid,
            "training_config": {
                "epochs_options": hypertune_config.epochs,
                "batch_size_options": hypertune_config.batch_sizes,
                "learning_rate_options": hypertune_config.learning_rates,
                "num_cpus_per_job": 2,
                "num_parallel_jobs": 4,
            },
            "model_config": {
                "model_framework": "pytorch",
                "model_config_name": f"Model {pipeline_config.id}",
                "model_type": "linear",
                "target_column_type": pipeline_config.target_type.value,
                "target_column": pipeline_config.target,
                "primary_metric": pipeline_config.primary_metric.value,
                "selected_features": list(pipeline_config.feature_configs.keys()),
                "loss_type": hypertune_config.loss_type.value,
                "advanced_options": advanced_options,
            },
            "next_model_id": next_model_id,
        },
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    api_status, response_json = extract_response("post_hypertune_job", response)
    if api_status == APIStatus.ERROR:
        return api_status, None
    return api_status, response_json["trained_model_metadata_uuids"]


def _parse_pipeline_config(
    pipeline_config_json: Dict[str, Any], update_dict: Dict[str, Any]
) -> PipelineConfig:
    """Parse a pipeline config from the SOTAI API.

    Args:
        pipeline_config_json: The pipeline config JSON to parse.
        update_dict: The dictionary with which to update the pipeline config.

    Returns:
        The parsed pipeline config.
    """
    pipeline_config = {}
    pipeline_config.update(pipeline_config_json)
    pipeline_config.update(update_dict)
    pipeline_config["dataset_split"] = DatasetSplit(
        train=pipeline_config_json["train_percentage"],
        val=pipeline_config_json["validation_percentage"],
        test=pipeline_config_json["test_percentage"],
    )
    feature_configs = {}
    for feature in pipeline_config_json["feature_config_list"]:
        feature["name"] = feature["feature_name"]
        feature["type"] = feature["feature_type"]
        if feature["feature_type"] == "categorical":
            if feature["categories_str"]:
                feature["categories"] = feature["categories_str"]
                del feature["categories_str"]
            else:
                feature["categories"] = feature["categories_int"]
                del feature["categories_int"]
            feature_configs[feature["name"]] = CategoricalFeatureConfig(**feature)
        else:
            feature_configs[feature["name"]] = NumericalFeatureConfig(**feature)

    pipeline_config["feature_configs"] = feature_configs
    pipeline_config["id"] = pipeline_config_json["pipeline_config_sdk_id"]
    pipeline_config["allow_hosting"] = True
    return PipelineConfig(**pipeline_config)


def get_pipeline(
    pipeline_uuid: str,
) -> Union[None, Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any], List[str]]]:
    """Get a pipeline from the SOTAI API.

    Args:
        pipeline_uuid: The UUID of the pipeline to get.

    Returns:
        A tuple containing the metadata for the pipeline, the id for the most recent
        config of the pipeline, the pipeline configs for the pipeline, and the UUIDs
        of the trainedmodels for the pipeline. If unsuccessful, returns `None`.
    """
    response = requests.get(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipelines/{pipeline_uuid}",
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    api_status, response_json = extract_response("get_pipeline", response)
    if api_status == APIStatus.ERROR:
        return None

    pipeline_metadata = {
        "name": response_json["name"],
        "target": response_json["target"],
        "target_type": response_json["target_type"],
        "primary_metric": response_json["primary_metric"],
        "shuffle_data": response_json["shuffle_data"],
        "drop_empty_percentage": response_json["drop_empty_percentage"],
    }

    last_config_id = -1
    pipeline_configs = {}

    for pipeline_config_json in response_json["pipeline_configs"]:
        pipeline_config = _parse_pipeline_config(
            pipeline_config_json, pipeline_metadata
        )
        last_config_id = max(last_config_id, pipeline_config.id)
        pipeline_configs[pipeline_config.id] = pipeline_config
    last_config = pipeline_configs[last_config_id]
    pipeline_metadata["dataset_split"] = last_config.dataset_split
    trained_model_uuids = response_json["trained_model_metadata_uuids"]
    return pipeline_metadata, last_config_id, pipeline_configs, trained_model_uuids


def get_trained_model_uuids(pipeline_uuid: str) -> List[str]:
    """Get the UUIDs of the trained models for a pipeline from the SOTAI API.

    Args:
        pipeline_uuid: The UUID of the pipeline to get the trained models for.

    Returns:
        The UUIDs of the trained models for the pipeline.
    """
    response = requests.get(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipelines/{pipeline_uuid}/trained-models",
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    api_status, _ = extract_response("get_trained_model_uuids", response)
    if api_status == APIStatus.ERROR:
        return []

    return [trained_model["uuid"] for trained_model in response.json()]


def get_trained_model_metadata(
    trained_model_uuid: str,
) -> Tuple[APIStatus, Optional[TrainedModelMetadata]]:
    """Get the metadata for a TrainedModelfrom the SOTAI API.

    Args:
        trained_model_uuid: The UUID of the trained model to get.

    Returns:
        The metadata for the trained model if training is complete, otherwise `None`.
    """
    response = requests.get(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/trained-model/{trained_model_uuid}",
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    _, response_json = extract_response("get_trained_model_metadata", response)
    if response_json["overall_model_results"] is None:
        return APIStatus.ERROR, None

    trained_model_metadata_json = response_json["trained_model_metadata"]
    overall_model_results = response_json["overall_model_results"]
    model_config = response_json["model_config"]
    pipeline_config_json = response_json["pipeline_config"]
    feature_analyses = response_json["feature_analyses"]
    trained_model_metadata = {
        "id": trained_model_metadata_json["trained_model_sdk_id"],
        "model_config": LinearConfig(
            output_calibration=model_config["output_calibration"],
            output_calibration_num_keypoints=model_config[
                "output_calibration_num_keypoints"
            ],
            output_initialization=model_config["output_initialization"],
            output_min=model_config["output_min"],
            output_max=model_config["output_max"],
            output_calibration_input_keypoints_type=model_config[
                "output_calibration_input_keypoints_type"
            ],
            use_bias=model_config["use_bias"],
        ),
        "training_config": TrainingConfig(
            epochs=trained_model_metadata_json["epochs"],
            batch_size=overall_model_results["batch_size"],
            learning_rate=overall_model_results["learning_rate"],
            loss_type=model_config["loss_type"],
        ),
        "training_results": TrainingResults(
            training_time=overall_model_results["runtime_in_seconds"],
            train_loss_by_epoch=overall_model_results["train_loss_per_epoch"],
            train_primary_metric_by_epoch=overall_model_results[
                "train_primary_metric_per_epoch"
            ],
            val_loss_by_epoch=overall_model_results["validation_loss_per_epoch"],
            val_primary_metric_by_epoch=overall_model_results[
                "validation_primary_metric_per_epoch"
            ],
            evaluation_time=overall_model_results["runtime_in_seconds"],
            test_loss=overall_model_results["test_loss"],
            test_primary_metric=overall_model_results["test_primary_metric"],
            feature_analyses={
                feature["feature_name"]: {
                    "feature_name": feature["feature_name"],
                    "feature_type": feature["feature_type"],
                    "min": feature["statistic_min"],
                    "max": feature["statistic_max"],
                    "mean": feature["statistic_mean"],
                    "median": feature["statistic_median"],
                    "std": feature["statistic_std"],
                    "keypoints_outputs": feature["keypoints_outputs"],
                    "keypoints_inputs_categorical": feature[
                        "keypoints_inputs_categorical"
                    ],
                    "keypoints_inputs_numerical": feature["keypoints_inputs_numerical"],
                }
                for feature in feature_analyses
            },
            linear_coefficients=dict(
                zip(
                    overall_model_results["feature_names"],
                    overall_model_results["linear_coefficients"],
                )
            ),
        ),
        "pipeline_config": _parse_pipeline_config(
            pipeline_config_json,
            {
                "target": model_config["target_column"],
                "target_type": model_config["target_column_type"],
                "primary_metric": model_config["primary_metric"],
            },
        ),
        "uuid": trained_model_uuid,
        "allow_hosting": True,
    }
    return APIStatus.SUCCESS, TrainedModelMetadata(**trained_model_metadata)


def download_trained_model(trained_model_uuid: str) -> str:
    """Download a trained model from the SOTAI API to a local tmp directory.

    Args:
        trained_model_uuid: The UUID of the trained model to download.

    Returns:
        The path to the downloaded model.
    """
    response = requests.get(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/trained-model/{trained_model_uuid}/download",
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    download_folder = f"/tmp/sotai/pipeline/trained_models/{trained_model_uuid}"
    download_path = f"{download_folder}/model.tar.gz"
    model_file_path = f"{download_folder}/model.pt"
    if not os.path.exists(download_folder):
        os.makedirs(download_folder)
    urllib.request.urlretrieve(response.json(), download_path)

    with tarfile.open(download_path) as model_file:
        model_file.extractall(download_folder)
    return model_file_path


def get_dataset_uuids(pipeline_uuid: str) -> Tuple[APIStatus, List[str]]:
    """Returns the UUIDs of the datasets for a pipeline from the SOTAI API.

    Args:
        pipeline_uuid: The UUID of the pipeline to get the datasets for.

    Returns:
        A tuple containing the status of the API call and the UUIDs of the datasets for
        the pipeline. If unsuccessful, the UUIDs will be `None`.
    """
    response = requests.get(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipelines/{pipeline_uuid}/datasets",
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    return extract_response("get_dataset_uuids", response)


def download_prepared_dataset(dataset_uuid: str) -> Tuple[APIStatus, Optional[Dataset]]:
    """Download a dataset from the SOTAI API.

    Args:
        trained_model_uuid: The UUID of the trained model to download.

    Returns:
        A tuple containing the status of the API call and the downloaded Dataset. If
        unsuccessful, the dataset will be `None`.
    """
    response = requests.get(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/datasets/{dataset_uuid}/download",
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    api_status, response_json = extract_response("download_prepared_dataset", response)
    if api_status == APIStatus.ERROR:
        return api_status, None

    download_folder = f"/tmp/sotai/pipeline/datasets/{dataset_uuid}"
    train_download_path = f"{download_folder}/train.csv"
    test_download_path = f"{download_folder}/test.csv"
    validation_download_path = f"{download_folder}/validation.csv"
    if not os.path.exists(download_folder):
        os.makedirs(download_folder)

    urllib.request.urlretrieve(response_json["train_download_url"], train_download_path)
    urllib.request.urlretrieve(response_json["test_download_url"], test_download_path)
    urllib.request.urlretrieve(
        response_json["validation_download_url"], validation_download_path
    )

    return APIStatus.SUCCESS, Dataset(
        prepared_data=PreparedData(
            train=pd.read_csv(train_download_path),
            val=pd.read_csv(validation_download_path),
            test=pd.read_csv(test_download_path),
        ),
        id=response_json["dataset_sdk_id"],
        pipeline_config_id=response_json["pipeline_config_sdk_id"],
        allow_hosting=True,
    )


def post_external_inference(
    shap_filepath: str,
    base_filepath: str,
    inference_filepath: str,
    beeswarm_filepath: str,
    scatter_filepath: str,
    feature_importance_filepath: str,
    external_shapley_value_name: str,
    target: str,
    dataset_name: str,
) -> Tuple[APIStatus, Optional[str]]:
    """Upload inference + shap values to the SOTAI API.

    Args:
        shap_filepath: The path to the shap values file to push to the API.
        base_filepath: The path to the base file to push to the API.
        inference_filepath: The path to the inference file to push to the API.
        external_shapley_value_name: The name of the external shapley value to create.
        target: The target column of the dataset.
        dataset_name: The name of the dataset.

    Returns:
        A tuple containing the status of the API call and the UUID of the created
        inference job. If unsuccessful, the UUID will be `None`.
    """

    response = requests.post(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/shapley-values",
        files=[
            # pylint: disable=consider-using-with
            ("files", open(shap_filepath, "rb")),
            (
                "files",
                open(base_filepath, "rb"),
            ),
            (
                "files",
                open(inference_filepath, "rb"),
            ),
            (
                "files",
                open(beeswarm_filepath, "rb"),
            ),
            (
                "files",
                open(scatter_filepath, "rb"),
            ),
            (
                "files",
                open(feature_importance_filepath, "rb"),
            ),
            # pylint: enable=consider-using-with
        ],
        data={
            "external_shapley_value_name": external_shapley_value_name,
            "target": target,
            "dataset_name": dataset_name,
        },
        headers=get_auth_headers(),
        timeout=SOTAI_API_TIMEOUT,
    )

    api_status, response_json = extract_response("post_external_inference", response)
    if api_status == APIStatus.ERROR:
        return api_status, None
    return api_status, response_json["uuid"]



"""Pydantic models for Pipelines."""
from typing import Dict, List, Literal, Optional, Union

import pandas as pd
from pydantic import BaseModel, Field, root_validator

from .enums import (
    FeatureType,
    HypertuneMethod,
    InputKeypointsInit,
    InputKeypointsType,
    LossType,
    Metric,
    Monotonicity,
    TargetType,
)


class DatasetSplit(BaseModel):
    """Defines the split percentage for train, val, and test datasets.

    Attributes:
        train: The percentage of the dataset to use for training.
        val: The percentage of the dataset to use for validation.
        test: The percentage of the dataset to use for testing.
    """

    train: int = 80
    val: int = 10
    test: int = 10

    @root_validator(pre=True, allow_reuse=True)
    @classmethod
    def validate_split_sum(cls, values):
        """Ensures that the split percentages add up to 100."""
        assert (
            values["train"] + values["val"] + values["test"] == 100
        ), "split percentages must add up to 100"
        return values


class PreparedData(BaseModel):
    """A train, val, and test set of data that's been cleaned.

    Attributes:
        train: The training dataset.
        val: The validation dataset.
        test: The testing dataset.
    """

    train: pd.DataFrame = Field(...)
    val: pd.DataFrame = Field(...)
    test: pd.DataFrame = Field(...)

    class Config:  # pylint: disable=missing-class-docstring,too-few-public-methods
        """Standard Pydantic BaseModel Config."""

        arbitrary_types_allowed = True


class Dataset(BaseModel):
    """A class for managing data.

    Attributes:
        id: The ID of the dataset used by the pipeline that prepared it.
        pipeline_config_id: The ID of the pipeline config used to create this dataset.
        prepared_data: The prepared data ready for training.
    """

    id: int = Field(...)
    pipeline_config_id: int = Field(...)
    prepared_data: PreparedData = Field(...)
    uuid: Optional[str] = None
    allow_hosting: bool = False

    class Config:  # pylint: disable=missing-class-docstring,too-few-public-methods
        """Standard Pydantic BaseModel Config."""

        arbitrary_types_allowed = True


class _BaseModelConfig(BaseModel):
    """Configuration for a calibrated model.

    Attributes:
        output_min: The minimum output value for the model. If None, then it will be
            assumed that there is no minimum output value.
        output_max: The maximum output value for the model. If None, then it will be
            assumed that there is no maximum output value.
        output_calibration: Whether to calibrate the output.
        output_calibration_num_keypoints: The number of keypoints to use for the output
            calibrator.
        output_initialization: The method for initializing the output calibrator input
            keypoints.
        output_calibration_input_keypoints_type: The type of output calibrator input
            keypoints.
    """

    output_min: Optional[float] = None
    output_max: Optional[float] = None
    output_calibration: bool = False
    output_calibration_num_keypoints: int = 10
    output_initialization: InputKeypointsInit = InputKeypointsInit.QUANTILES
    output_calibration_input_keypoints_type: InputKeypointsType = (
        InputKeypointsType.FIXED
    )


class LinearConfig(_BaseModelConfig):
    """Configuration for a calibrated linear model.

    Attributes:
        use_bias: Whether to use a bias term for the linear combination.
    """

    use_bias: bool = True


class TrainingConfig(BaseModel):
    """Configuration for training a single model.

    Attributes:
        loss_type: The type of loss function to use for training.
        epochs: The number of iterations through the dataset during training.
        batch_size: The number of examples to use for each training step.
        learning_rate: The learning rate to use for the optimizer.
    """

    loss_type: LossType = Field(...)
    epochs: int = 100
    batch_size: int = 32
    learning_rate: float = 1e-3


class FeatureAnalysis(BaseModel):
    """Feature analysis results for a single feature of a trained model.

    Attributes:
        feature_name: The name of the feature.
        feature_type: The type of the feature.
        min: The minimum value of the feature.
        max: The maximum value of the feature.
        mean: The mean value of the feature.
        median: The median value of the feature.
        std: The standard deviation of the feature.
        keypoints_inputs_numerical: The input keypoints for the feature if the feature
            is numerical. Otherwise this should be `None`.
        keypoints_inputs_categorical: The input keypoints for the feature if the feature
            is categorical. Otherwise this should be `None`.
        keypoints_outputs: The output keypoints for each input keypoint.
    """

    feature_name: str = Field(...)
    feature_type: FeatureType = Field(...)
    min: Optional[float] = None
    max: Optional[float] = None
    mean: Optional[float] = None
    median: Optional[float] = None
    std: Optional[float] = None
    # One of the keypoint inputs must exist, which one depends on feature_type.
    keypoints_inputs_numerical: Optional[List[float]] = Field(...)
    keypoints_inputs_categorical: Optional[List[str]] = Field(...)
    keypoints_outputs: List[float] = Field(...)


class TrainingResults(BaseModel):
    """Training results for a single calibrated model.

    Attributes:
        training_time: The total time spent training the model.
        train_loss_by_epoch: The training loss for each epoch.
        train_primary_metric_by_epoch: The training primary metric for each epoch.
        val_loss_by_epoch: The validation loss for each epoch.
        val_primary_metric_by_epoch: The validation primary metric for each
            epoch.
        evaluation_time: The total time spent evaluating the model.
        test_loss: The test loss.
        test_primary_metric: The test primary metric.
        feature_analyses: The feature analysis results for each feature.
        linear_coefficients: A mapping from feature name to linear coefficient. These
            coefficients are the coefficients of the linear combination of features
            after they have been calibrated, so any analysis of the coefficients should
            be done with the feature's calibrator in mind. If using a bias term, the
            bias value will be stored under the key "bias". Note that there will be no
            bias term if you set output bounds or use an output calibrator.
    """

    training_time: float = Field(...)
    train_loss_by_epoch: List[float] = Field(...)
    train_primary_metric_by_epoch: List[float] = Field(...)
    val_loss_by_epoch: List[float] = Field(...)
    val_primary_metric_by_epoch: List[float] = Field(...)
    evaluation_time: float = Field(...)
    test_loss: float = Field(...)
    test_primary_metric: float = Field(...)
    feature_analyses: Dict[str, FeatureAnalysis] = Field(...)
    linear_coefficients: Dict[str, float] = Field(...)


class NumericalFeatureConfig(BaseModel):
    """Configuration for a numerical feature.

    Attributes:
        name: The name of the feature.
        type: The type of the feature. Always `FeatureType.NUMERICAL`.
        num_keypoints: The number of keypoints to use for the calibrator.
        input_keypoints_init: The method for initializing the input keypoints.
        input_keypoints_type: The type of input keypoints.
        monotonicity: The monotonicity constraint, if any.
        projection_iterations: Number of times to run Dykstra's projection algorithm
            when applying constraints.
    """

    name: str = Field(...)
    type: Literal[FeatureType.NUMERICAL] = Field(FeatureType.NUMERICAL, frozen=True)
    num_keypoints: int = 10
    input_keypoints_init: InputKeypointsInit = InputKeypointsInit.QUANTILES
    input_keypoints_type: InputKeypointsType = InputKeypointsType.FIXED
    monotonicity: Monotonicity = Monotonicity.NONE
    projection_iterations: int = 8


class CategoricalFeatureConfig(BaseModel):
    """Configuration for a categorical feature.

    Attributes:
        name: The name of the feature.
        type: The type of the feature. Always `FeatureType.CATEGORICAL`.
        categories: The categories for the feature.
    """

    name: str = Field(...)
    type: Literal[FeatureType.CATEGORICAL] = Field(FeatureType.CATEGORICAL, frozen=True)
    categories: Union[List[int], List[str]] = Field(...)
    # TODO (will): add support for categorical monotonicity.


class PipelineConfig(BaseModel):
    """A configuration object for a `Pipeline`.

    Attributes:
        id: The ID of the pipeline config. This will be set by the Pipeline when it
            versions this config during preparation.
        uuid: The UUID of the pipeline.
        target: The column name for the target.
        target_type: The type of the target.
        primary_metric: The primary metric to use for training and evaluation.
        feature_configs: A dictionary mapping the column name for a feature to its
            config.
        shuffle_data: Whether to shuffle the data before splitting it into train,
            validation, and test sets.
        drop_empty_percentage: Rows will be dropped if they are this percentage empty.
        dataset_split: The split of the dataset into train, validation, and test sets.
    """

    id: Optional[int] = None
    uuid: Optional[str] = None
    target: str = Field(...)
    target_type: TargetType = Field(...)
    primary_metric: Metric = Field(...)
    feature_configs: Dict[
        str, Union[CategoricalFeatureConfig, NumericalFeatureConfig]
    ] = Field(...)
    shuffle_data: bool = Field(...)
    drop_empty_percentage: int = Field(...)
    dataset_split: DatasetSplit = Field(...)
    allow_hosting: bool = False


class HypertuneConfig(BaseModel):
    """A configuration object for hypertuning.

    Attributes:
        loss_type: The type of loss function to use for training.
        epochs: A list of epochs to try.
        batch_sizes: A list of batch sizes to try.
        learning_rates: A list of learning rates to try.
        hypertune_method: The method to use for hypertuning.
    """

    loss_type: LossType
    epochs: List[int]
    batch_sizes: List[int]
    learning_rates: List[float]
    hypertune_method: HypertuneMethod = Field(default=HypertuneMethod.GRID)


class TrainedModelMetadata(BaseModel):
    """Metadata for a trained model.

    Attributes:
        id: The ID of the trained model.
        dataset_id: The ID of the dataset used to train the model.
        pipeline_uuid: The UUID of the pipeline used to train the model. This will be
        `None` if the trained model has not been analyzed under a pipeline.
        pipeline_config: The configuration of the pipeline used to train the model.
        model_config: The configuration of the model used to train the model.
        training_config: The training configuration used to train the model.
        training_results: The results of training the model.
        uuid: The UUID of the trained model. This will be `None` if the trained model
        has not been analyzed under a pipeline.
        analysis_url: The URL of the analysis of the trained model. This will be `None`
        if the trained model has not been analyzed under a pipeline.
    """

    id: Optional[int] = None
    dataset_id: Optional[int] = None
    pipeline_uuid: Optional[str] = None
    pipeline_config: PipelineConfig = Field(...)
    # Note: model_config is protected in Pydantic >=2.0.0, so we use <2.0.0
    model_config: LinearConfig = Field(...)
    training_config: TrainingConfig = Field(...)
    training_results: TrainingResults = Field(...)
    uuid: Optional[str] = None
    analysis_url: Optional[str] = None
    allow_hosting: bool = False



"""Lattice module for use in calibrated modeling.

PyTorch implementation of a lattice layer.
This layer takes one or more d-dimensional inputs and outputs the interpolated value
according the specified interpolation method.
"""
from typing import Iterator, Tuple, Union, List, Callable

import numpy as np
import torch

from ..enums import Interpolation, LatticeInit, Monotonicity


# pylint: disable=too-many-instance-attributes
class Lattice(torch.nn.Module):
    """A Lattice Module.

    Layer performs interpolation using one of 'units' d-dimensional lattices with
    arbitrary number of keypoints per dimension. Each lattice vertex has a trainable
    weight, and input is considered to be a d-dimensional point within the lattice.

    Attributes:
      - All `__init__` arguments.
      kernel: `torch.nn.Parameter` of shape `(prod(lattice_sizes), units)` which stores
        weights at each vertex of lattice.

    Example:
    `python
    lattice_sizes = [2, 2, 4, 3]
    inputs=torch.tensor(...) # shape: (batch_size, len(lattice_sizes))
    lattice=Lattice(
      lattice_sizes,
      clip_inputs=True,
      interpolation=Interpolation.HYPERCUBE,
      units=1,
    )
    outputs = Lattice(inputs)
    `
    """

    def __init__(
        self,
        lattice_sizes: Union[List[int], Tuple[int]],
        output_min: float = 0.0,
        output_max: float = 1.0,
        kernel_init: LatticeInit = LatticeInit.LINEAR,
        clip_inputs: bool = True,
        interpolation: Interpolation = Interpolation.HYPERCUBE,
        units: int = 1,
    ) -> None:
        """Initializes an instance of 'Lattice'.

        Args:
            lattice_sizes: List or tuple of size of lattice along each dimension.
            output_min: Minimum output value for weights at vertices of lattice.
            output_max: Maximum output value for weights at vertices of lattice.
            kernel_init: Initialization scheme to use for the kernel.
            clip_inputs: Whether input points should be clipped to the range of lattice.
            interpolation: Interpolation scheme for a given input.
            units: Dimensionality of weights stored at each vertex of lattice.

        Raises:
            ValueError: if `kernel_init` is invalid.
            NotImplementedError: Random monotonic initialization not yet implemented.
        """
        super().__init__()

        self.lattice_sizes = list(lattice_sizes)
        self.output_min = output_min
        self.output_max = output_max
        self.kernel_init = kernel_init
        self.clip_inputs = clip_inputs
        self.interpolation = interpolation
        self.units = units

        @torch.no_grad()
        def initialize_kernel() -> torch.Tensor:
            if self.kernel_init == LatticeInit.LINEAR:
                return self._linear_initializer()
            if self.kernel_init == LatticeInit.RANDOM_MONOTONIC:
                raise NotImplementedError(
                    "Random monotonic initialization not yet implemented."
                )
            raise ValueError(f"Unknown kernel init: {self.kernel_init}")

        self.kernel = torch.nn.Parameter(initialize_kernel())

    # pylint: disable-next=invalid-name
    def forward(self, x: Union[torch.Tensor, List[torch.Tensor]]) -> torch.Tensor:
        """Calculates interpolation from input, using method of self.interpolation.

        Args:
            x: input tensor. If `units == 1`, tensor of shape:
              `(batch_size, ..., len(lattice_size))` or list of `len(lattice_sizes)`
              tensors of same shape: `(batch_size, ..., 1)`. If `units > 1`, tensor of
              shape `(batch_size, ..., units, len(lattice_sizes))` or list of
              `len(lattice_sizes)` tensors of same shape `(batch_size, ..., units, 1)`.

        Returns:
            torch.Tensor of shape `(batch_size, ..., units)` containing interpolated
            values.

        Raises:
            ValueError: If the type of interpolation is unknown.
        """
        if self.interpolation == Interpolation.HYPERCUBE:
            return self._compute_hypercube_interpolation(x.double())
        if self.interpolation == Interpolation.SIMPLEX:
            return self._compute_simplex_interpolation(x.double())
        raise ValueError(f"Unknown interpolation type: {self.interpolation}")

    ################################################################################
    ############################## PRIVATE METHODS #################################
    ################################################################################

    def _linear_initializer(
        self,
        monotonicities: List[Monotonicity] = None,
        unimodalities=None,
    ) -> torch.Tensor:
        """Creates initial weights tensor for linear initialization.

        Args:
            monotonicities: monotonicity constraints of lattice, enforced in
              initialization.
            unimodalities: unimodality constraints of lattice, enforced in
              initialization.

        Returns:
            `torch.Tensor` of shape `(prod(lattice_sizes), units)`
        """
        # TODO: convert counting logic of monotoncities/unimodalities to suit enums.
        if monotonicities is None:
            monotonicities = [0] * len(self.lattice_sizes)
        if unimodalities is None:
            unimodalities = [0] * len(self.lattice_sizes)

        num_constraint_dims = self._count_non_zeros(monotonicities, unimodalities)

        if num_constraint_dims == 0:
            monotonicities = [1] * len(self.lattice_sizes)
            num_constraint_dims = len(self.lattice_sizes)

        dim_range = float(self.output_max - self.output_min) / num_constraint_dims
        one_d_weights = []

        for monotonicity, unimodality, dim_size in zip(
            monotonicities, unimodalities, self.lattice_sizes
        ):
            if monotonicity != 0:
                one_d = np.linspace(start=0.0, stop=dim_range, num=dim_size)
            elif unimodality != 0:
                decreasing = np.linspace(
                    start=dim_range, stop=0.0, num=(dim_size + 1) // 2
                )
                increasing = np.linspace(
                    start=0.0, stop=dim_range, num=(dim_size + 1) // 2
                )
                if unimodality == 1:
                    one_d = np.concatenate((decreasing, increasing[dim_size % 2 :]))
                else:
                    one_d = np.concatenate((increasing, decreasing[dim_size % 2 :]))
            else:
                one_d = np.array([0.0] * dim_size)

            one_d_weights.append(torch.tensor(one_d, dtype=torch.double).unsqueeze(0))

        weights = self._batch_outer_operation(one_d_weights, operation=torch.add)
        weights = (weights + self.output_min).view(-1, 1)
        if self.units > 1:
            weights = weights.repeat(1, self.units)

        return weights

    @staticmethod
    def _count_non_zeros(*iterables) -> int:
        """Returns total number of non 0 elements in given iterables.

        Args:
            *iterables: Any number of the value `None` or iterables of numeric values.
        """
        result = 0
        for iterable in iterables:
            if iterable is not None:
                result += sum(1 for element in iterable if element != 0)
        return result

    # pylint: disable=too-many-locals
    def _compute_simplex_interpolation(self, inputs: torch.Tensor) -> torch.Tensor:
        """Evaluates a lattice using simplex interpolation.

        Each `d`-dimensional unit hypercube of the lattice can be partitioned into `d!`
        disjoint simplices with `d+1` vertices. `S` is the unique simplex which contains
        input point `P`, and `S` has vertices `ABCD...`. For any vertex such as `A`, a
        new simplex `S'` can be created using the vertices `PBCD...`. The weight of `A`
        within the interpolation is then `vol(S')/vol(S)`. This process is repeated
        for every vertex in `S`, and the resulting values are summed.

        This interpolation can be computed in `O(D log(D))` time because it is only
        necessary to compute the volume of the simplex containing input point `P`. For
        context, the unit hypercube can be partitioned into `d!` simplices by starting
        at `(0,0,...,0)` and incrementing `0` to `1` dimension-by-dimensionuntil one
        reaches `(1,1,...,1)`. There are `d!` possible paths from `(0,0,...,0)` to
        `(1,1,...,1)`, which account for the number of unique, disjoint simplices
        created by the method. There are `d` steps for each possible path where each
        step comprises the vertices of one simplex. Thus, one can find the containing
        simplex for input `P` by argsorting the coordinates of `P` in descending order
        and pathing along said order. To compute the intepolation weights simply take
        the deltas from `[1, desc_sort(P_coords), 0]`.

        Args:
            inputs: input tensor. If `units == 1`, tensor of shape:
              `(batch_size, ..., len(lattice_size))` or list of `len(lattice_sizes)`
              tensors of same shape: `(batch_size, ..., 1)`. If `units > 1`, tensor of
              shape `(batch_size, ..., units, len(lattice_sizes))` or list of
              `len(lattice_sizes)` tensors of same shape `(batch_size, ..., units, 1)`

        Returns:
            `torch.Tensor` of shape `(batch_size, ..., units)` containing interpolated
            values.
        """
        if isinstance(inputs, list):
            inputs = torch.cat(inputs, dim=-1)

        if self.clip_inputs:
            inputs = self._clip_onto_lattice_range(inputs)

        lattice_rank = len(self.lattice_sizes)
        input_dim = len(inputs.shape)
        all_size_2 = all(size == 2 for size in self.lattice_sizes)

        # Strides are the index shift (with respect to flattened kernel data) of each
        # dimension, which can be used in a dot product with multi-dimensional
        # coordinates to give an index for the flattened lattice weights.
        # Ex): for lattice_sizes = [4, 3, 2], we get strides = [6, 2, 1]: when looking
        # at lattice coords (i, j, k) and kernel data flattened into 1-D, incrementing i
        # corresponds to a shift of 6 in flattened kernel data, j corresponds to a shift
        # of 2, and k corresponds to a shift of 1. Consequently, we can do
        # (coords * strides) for any coordinates to obtain the flattened index.
        strides = torch.tensor(
            np.cumprod([1] + self.lattice_sizes[::-1][:-1])[::-1].copy()
        )
        lower_corner_offset = None
        if not all_size_2:
            lower_corner_coordinates = inputs.int()
            lower_corner_coordinates = torch.min(
                lower_corner_coordinates, torch.tensor(self.lattice_sizes) - 2
            )
            lower_corner_offset = (lower_corner_coordinates * strides).sum(
                dim=-1, keepdim=True
            )
            inputs = inputs - lower_corner_coordinates.float()

        sorted_indices = torch.argsort(inputs, descending=True)
        sorted_inputs = torch.sort(inputs, descending=True).values

        # Pad the 1 and 0 onto the ends of sorted coordinates and compute deltas.
        no_padding_dims = [(0, 0)] * (input_dim - 1)
        flat_no_padding = [item for sublist in no_padding_dims for item in sublist]
        sorted_inputs_padded_left = torch.nn.functional.pad(
            sorted_inputs, [1, 0] + flat_no_padding, value=1.0
        )
        sorted_inputs_padded_right = torch.nn.functional.pad(
            sorted_inputs, [0, 1] + flat_no_padding, value=0.0
        )
        weights = sorted_inputs_padded_left - sorted_inputs_padded_right

        # Use strides to find indices of simplex vertices in flattened form.
        sorted_strides = torch.gather(strides, 0, sorted_indices.view(-1)).view(
            sorted_indices.shape
        )
        if all_size_2:
            corner_offset_and_sorted_strides = torch.nn.functional.pad(
                sorted_strides, [1, 0] + flat_no_padding
            )
        else:
            corner_offset_and_sorted_strides = torch.cat(
                [lower_corner_offset, sorted_strides], dim=-1
            )
        indices = torch.cumsum(corner_offset_and_sorted_strides, dim=-1)

        # Get kernel data from corresponding simplex vertices.
        if self.units == 1:
            gathered_params = torch.index_select(
                self.kernel.view(-1), 0, indices.view(-1)
            ).view(indices.shape)
        else:
            unit_offset = torch.tensor(
                [[i] * (lattice_rank + 1) for i in range(self.units)]
            )
            flat_indices = indices * self.units + unit_offset
            gathered_params = torch.index_select(
                self.kernel.view(-1), 0, flat_indices.view(-1)
            ).view(indices.shape)

        return (gathered_params * weights).sum(dim=-1, keepdim=self.units == 1)

    # pylint: enable=too-many-locals

    def _compute_hypercube_interpolation(
        self,
        inputs: Union[torch.Tensor, List[torch.Tensor]],
    ) -> torch.Tensor:
        """Performs hypercube interpolation using the surrounding unit hypercube.

        Args:
            inputs: input tensor. If `units == 1`, tensor of shape:
              `(batch_size, ..., len(lattice_size))` or list of `len(lattice_sizes)`
              tensors of same shape: `(batch_size, ..., 1)`. If `units > 1`, tensor of
              shape `(batch_size, ..., units, len(lattice_sizes))` or list of
              `len(lattice_sizes)` tensors of same shape `(batch_size, ..., units, 1)`

        Returns:
            `torch.Tensor` of shape `(batch_size, ..., units)` containing interpolated
            value(s).
        """
        interpolation_weights = self._compute_hypercube_interpolation_weights(
            inputs=inputs, clip_inputs=self.clip_inputs
        )
        if self.units == 1:
            return torch.matmul(interpolation_weights, self.kernel)

        return torch.sum(interpolation_weights * self.kernel.t(), dim=-1)

    def _compute_hypercube_interpolation_weights(
        self, inputs: Union[torch.Tensor, List[torch.Tensor]], clip_inputs: bool = True
    ) -> torch.Tensor:
        """Computes weights for hypercube lattice interpolation.

        For each n-dim unit in "inputs," the weights matrix will generate the weights
        corresponding to the unit's location within its surrounding hypercube. These
        weights can then be multiplied by the lattice layer's kernel to compute the
        actual hypercube interpolation. Specifically, the outer product of the set
        `(1-x_i, x_i)` for all x_i in input unit x calculates the weights for each
        vertex in the surrounding hypercube, and every other vertex in the lattice is
        set to zero since it is not used. In addition, for consecutive dimensions of
        equal size in the lattice, broadcasting is used to speed up calculations.

        Args:
            inputs: torch.Tensor of shape `(batch_size, ..., len(lattice_sizes)` or list
              of `len(lattice_sizes)` tensors of same shape `(batch_size, ..., 1)`
            clip_inputs: Boolean to determine whether input values outside lattice
              bounds should be clipped to the min or max supported values.

        Returns:
            `torch.Tensor` of shape `(batch_size, ..., prod(lattice_sizes))` containing
            the weights which can be matrix multiplied with the kernel to perform
            hypercube interpolation.
        """
        if isinstance(inputs, list):
            input_dtype = inputs[0].dtype
        else:
            input_dtype = inputs.dtype

        # Special case: 2^d lattice with input passed in as a single tensor
        if all(size == 2 for size in self.lattice_sizes) and not isinstance(
            inputs, list
        ):
            # pylint: disable=invalid-name
            w = torch.stack([(1.0 - inputs), inputs], dim=-1)
            if clip_inputs:
                w = torch.clamp(w, min=0, max=1)
            # pylint: enable=invalid-name
            one_d_interpolation_weights = list(torch.unbind(w, dim=-2))
            return self._batch_outer_operation(one_d_interpolation_weights)

        if clip_inputs:
            inputs = self._clip_onto_lattice_range(inputs)

        # Set up buckets of consecutive equal dimensions for broadcasting later
        dim_keypoints = {}
        for dim_size in set(self.lattice_sizes):
            dim_keypoints[dim_size] = torch.tensor(
                list(range(dim_size)), dtype=input_dtype
            )
        bucketized_inputs = self._bucketize_consecutive_equal_dims(inputs)
        one_d_interpolation_weights = []

        for tensor, bucket_size, dim_size in bucketized_inputs:
            if bucket_size > 1:
                tensor = torch.unsqueeze(tensor, dim=-1)
            distance = torch.abs(tensor - dim_keypoints[dim_size])
            weights = 1.0 - torch.minimum(
                distance, torch.tensor(1.0, dtype=distance.dtype)
            )
            if bucket_size == 1:
                one_d_interpolation_weights.append(weights)
            else:
                one_d_interpolation_weights.extend(torch.unbind(weights, dim=-2))

        return self._batch_outer_operation(one_d_interpolation_weights)

    @staticmethod
    def _batch_outer_operation(
        list_of_tensors: List[torch.Tensor],
        operation: Union[str, Callable] = "auto",
    ) -> torch.Tensor:
        """Computes the flattened outer product of a list of tensors.

        Args:
            list_of_tensors: List of tensors of same shape `(batch_size, ..., k[i])`
              where everything except `k_i` matches.

        Returns:
            `torch.Tensor` of shape `(batch_size, ..., k_i * k_j * ...)` containing a
            flattened version of the outer product.
        """
        if len(list_of_tensors) == 1:
            return list_of_tensors[0]

        result = torch.unsqueeze(list_of_tensors[0], dim=-1)

        for i, tensor in enumerate(list_of_tensors[1:]):
            # pylint: disable=invalid-name
            if operation == "auto":
                op = torch.mul if i < 6 else torch.matmul
            else:
                op = operation
            # pylint: enable=invalid-name

            result = op(result, torch.unsqueeze(tensor, dim=-2))
            shape = [-1] + [int(size) for size in result.shape[1:]]
            new_shape = shape[:-2] + [shape[-2] * shape[-1]]
            if i < len(list_of_tensors) - 2:
                new_shape.append(1)
            result = torch.reshape(result, new_shape)

        return result

    def _clip_onto_lattice_range(
        self,
        inputs: Union[torch.Tensor, List[torch.Tensor]],
    ) -> torch.Tensor:
        """Clips inputs onto valid input range for given lattice_sizes.

        Args:
            inputs: `inputs` argument of `_compute_interpolation_weights()`.

        Returns:
            `torch.Tensor` of shape `inputs` with values within range
            `[0, dim_size - 1]`.
        """
        if not isinstance(inputs, list):
            upper_bounds = torch.tensor(
                [dim_size - 1.0 for dim_size in self.lattice_sizes]
            ).double()
            clipped_inputs = torch.clamp(
                inputs, min=torch.zeros_like(upper_bounds), max=upper_bounds
            )
        else:
            dim_upper_bounds = {}
            for dim_size in set(self.lattice_sizes):
                dim_upper_bounds[dim_size] = torch.tensor(
                    dim_size - 1.0, dtype=inputs[0].dtype
                )
            dim_lower_bound = torch.zeros(1, dtype=inputs[0].dtype)

            clipped_inputs = []
            for one_d_input, dim_size in zip(inputs, self.lattice_sizes):
                clipped_inputs.append(
                    torch.clamp(
                        one_d_input, min=dim_lower_bound, max=dim_upper_bounds[dim_size]
                    )
                )

        return clipped_inputs

    def _bucketize_consecutive_equal_dims(
        self,
        inputs: Union[torch.Tensor, List[torch.Tensor]],
    ) -> Iterator[Tuple[torch.Tensor, int, int]]:
        """Creates buckets of equal sized dimensions for broadcasting ops.

        Args:
            inputs: `inputs` argument of `_compute_interpolation_weights()`.

        Returns:
            An `Iterable` containing `(torch.Tensor, int, int)` where the tensor
            contains individual values from "inputs" corresponding to its bucket, the
            first `int` is bucket size, and the second `int` is size of the dimension of
            the bucket.
        """
        if not isinstance(inputs, list):
            bucket_sizes = []
            bucket_dim_sizes = []
            current_size = 1
            for i in range(1, len(self.lattice_sizes)):
                if self.lattice_sizes[i] != self.lattice_sizes[i - 1]:
                    bucket_sizes.append(current_size)
                    bucket_dim_sizes.append(self.lattice_sizes[i - 1])
                    current_size = 1
                else:
                    current_size += 1
            bucket_sizes.append(current_size)
            bucket_dim_sizes.append(self.lattice_sizes[-1])
            inputs = torch.split(inputs, split_size_or_sections=bucket_sizes, dim=-1)
        else:
            bucket_sizes = [1] * len(self.lattice_sizes)
            bucket_dim_sizes = self.lattice_sizes

        return zip(inputs, bucket_sizes, bucket_dim_sizes)



"""This module contains functions for external models to interact with the SOTAI API."""
import os
import pickle
import numpy as np
import pandas as pd
from .utils.shap_utils import (
    calculate_beeswarm,
    calculate_scatter,
    calculate_feature_importance,
)
from .api import post_external_inference


def shap(  # pylint: disable=too-many-locals
    inference_data: pd.DataFrame,
    shapley_values: np.ndarray,
    base_values: np.ndarray,
    name: str,
    target: str,
    dataset_name: str,
) -> str:
    """Uploads the shapley values, base values, and inference data to the SOTAI API.

    Args:
        inference_data: The data used for inference.
        shapley_values: The shapley values for the inference data.
        base_values: The base values for the inference data.
        name: The name of the shapley values.
        target: The target column of the inference data.
        dataset_name: The name of the dataset used for inference.

    Returns:
        The UUID of the uploaded shapley values.
    """

    external_directory = "/tmp/sotai/external/"
    if not os.path.exists(external_directory):
        os.makedirs(external_directory)

    beeswarm_data = calculate_beeswarm(inference_data, shapley_values, target)
    scatter_data = calculate_scatter(inference_data, shapley_values)
    feature_importance = calculate_feature_importance(
        shapley_values, inference_data.columns
    )

    beeswarm_filepath = "/tmp/sotai/external/beeswarm_data.pkl"
    scatter_filepath = "/tmp/sotai/external/scatter_data.pkl"
    feature_importance_filepath = "/tmp/sotai/external/feature_importance_data.pkl"

    with open(beeswarm_filepath, "wb") as beeswarm_file:
        pickle.dump(beeswarm_data, beeswarm_file)
    with open(scatter_filepath, "wb") as scatter_file:
        pickle.dump(scatter_data, scatter_file)
    with open(feature_importance_filepath, "wb") as feature_importance_file:
        pickle.dump(feature_importance, feature_importance_file)

    shapley_values_df = pd.DataFrame(shapley_values)
    base_values_df = pd.DataFrame(base_values)

    shapley_value_filepath = "/tmp/sotai/external/shapley_values.csv"
    base_values_filepath = "/tmp/sotai/external/base_values.csv"
    inference_data_filepath = "/tmp/sotai/external/inference_predictions.csv"

    shapley_values_df.to_csv(shapley_value_filepath, index=False)
    base_values_df.to_csv(
        base_values_filepath,
        index=False,
    )
    inference_data.to_csv(inference_data_filepath, index=False)

    shap_uuid = post_external_inference(
        external_shapley_value_name=name,
        shap_filepath=shapley_value_filepath,
        base_filepath=base_values_filepath,
        inference_filepath=inference_data_filepath,
        beeswarm_filepath=beeswarm_filepath,
        scatter_filepath=scatter_filepath,
        feature_importance_filepath=feature_importance_filepath,
        target=target,
        dataset_name=dataset_name,
    )
    return shap_uuid



"""Numerical calibration module.

PyTorch implementation of the numerical calibration module. This module takes in a
single-dimensional input and transforms it using piece-wise linear functions that
satisfy desired bounds and monotonicity constraints.
"""
from collections import defaultdict
from typing import List, Optional, Tuple

import numpy as np
import torch

from ..enums import Monotonicity, NumericalCalibratorInit


# pylint: disable-next=too-many-instance-attributes
class NumericalCalibrator(torch.nn.Module):
    """A numerical calibrator.

    This module takes an input of shape `(batch_size, 1)` and calibrates it using a
    piece-wise linear function that conforms to any provided constraints. The output
    will have the same shape as the input.

    Attributes:
        - All `__init__` arguments.
        kernel: `torch.nn.Parameter` that stores the piece-wise linear function weights.
        missing_output: `torch.nn.Parameter` that stores the output learned for any
            missing inputs. Only available if `missing_input_value` is provided.

    Example:
    ```python
    inputs = torch.tensor(...)  # shape: (batch_size, 1)
    calibrator = NumericalCalibrator(
        input_keypoints=np.linspace(1., 5., num=5),
        output_min=0.0,
        output_max=1.0,
        monotonicity=Monotonicity.INCREASING,
        kernel_init=NumericalCalibratorInit.EQUAL_HEIGHTS,
    )
    outputs = calibrator(inputs)
    ```
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        input_keypoints: np.ndarray,
        missing_input_value: Optional[float] = None,
        output_min: Optional[float] = None,
        output_max: Optional[float] = None,
        monotonicity: Monotonicity = Monotonicity.NONE,
        kernel_init: NumericalCalibratorInit = NumericalCalibratorInit.EQUAL_HEIGHTS,
        projection_iterations: int = 8,
    ) -> None:
        """Initializes an instance of `NumericalCalibrator`.

        Args:
            input_keypoints: Ordered list of float-valued keypoints for the underlying
                piece-wise linear function.
            missing_input_value: If provided, the calibrator will learn to map all
                instances of this missing input value to a learned output value.
            output_min: Minimum output value. If `None`, the minimum output value will
                be unbounded.
            output_max: Maximum output value. If `None`, the maximum output value will
                be unbounded.
            monotonicity: Monotonicity constraint for the underlying piece-wise linear
                function.
            kernel_init: Initialization scheme to use for the kernel.
            projectionion_iterations: Number of times to run Dykstra's projection
                algorithm when applying constraints.

        Raises:
            ValueError: If `kernel_init` is invalid.
        """
        super().__init__()

        self.input_keypoints = input_keypoints
        self.missing_input_value = missing_input_value
        self.output_min = output_min
        self.output_max = output_max
        self.monotonicity = monotonicity
        self.kernel_init = kernel_init
        self.projection_iterations = projection_iterations

        # Determine default output initialization values if bounds are not fully set.
        if output_min is not None and output_max is not None:
            output_init_min, output_init_max = self.output_min, self.output_max
        elif output_min is not None:
            output_init_min, output_init_max = self.output_min, self.output_min + 4.0
        elif output_max is not None:
            output_init_min, output_init_max = self.output_max - 4.0, self.output_max
        else:
            output_init_min, output_init_max = -2.0, 2.0
        self._output_init_min, self._output_init_max = output_init_min, output_init_max

        self._interpolation_keypoints = torch.from_numpy(input_keypoints[:-1])
        self._lengths = torch.from_numpy(input_keypoints[1:] - input_keypoints[:-1])

        # First row of the kernel represents the bias. The remaining rows represent
        # the y-value delta compared to the previous point i.e. the segment heights.
        @torch.no_grad()
        def initialize_kernel() -> torch.Tensor:
            output_init_range = self._output_init_max - self._output_init_min
            if kernel_init == NumericalCalibratorInit.EQUAL_HEIGHTS:
                num_segments = self._interpolation_keypoints.size()[0]
                segment_height = output_init_range / num_segments
                heights = torch.tensor([[segment_height]] * num_segments)
            elif kernel_init == NumericalCalibratorInit.EQUAL_SLOPES:
                heights = (
                    self._lengths * output_init_range / torch.sum(self._lengths)
                )[:, None]
            else:
                raise ValueError(f"Unknown kernel init: {self.kernel_init}")

            if monotonicity == Monotonicity.DECREASING:
                bias = self._output_init_max
                heights = -heights
            else:
                bias = self._output_init_min
            bias = torch.tensor([[bias]])
            return torch.cat((bias, heights), 0).double()

        self.kernel = torch.nn.Parameter(initialize_kernel())

        if missing_input_value:
            self.missing_output = torch.nn.Parameter(torch.Tensor(1))
            torch.nn.init.constant_(
                self.missing_output,
                (self._output_init_min + self._output_init_max) / 2.0,
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # pylint: disable=invalid-name
        """Calibrates numerical inputs through piece-wise linear interpolation.

        Args:
            x: The input tensor of shape `(batch_size, 1)`.

        Returns:
            torch.Tensor of shape `(batch_size, 1)` containing calibrated input values.
        """
        interpolation_weights = (x - self._interpolation_keypoints) / self._lengths
        interpolation_weights = torch.minimum(interpolation_weights, torch.tensor(1.0))
        interpolation_weights = torch.maximum(interpolation_weights, torch.tensor(0.0))
        interpolation_weights = torch.cat(
            (torch.ones_like(x), interpolation_weights), -1
        )
        result = torch.mm(interpolation_weights, self.kernel)

        if self.missing_input_value is not None:
            missing_mask = torch.eq(x, self.missing_input_value).long()
            result = missing_mask * self.missing_output + (1.0 - missing_mask) * result

        return result

    @torch.no_grad()
    def assert_constraints(self, eps=1e-6) -> List[str]:
        """Asserts that layer satisfies specified constraints.

        This checks that weights follow the layer's monotonicity and that the output is
        within bounds.

        Args:
            eps: the margin of error allowed

        Returns:
            A list of messages describing violated constraints including indices of
            monotonicity violations. If no constraints violated, the list will be empty.
        """
        weights = torch.squeeze(self.kernel.data)
        messages = []

        if self.output_max is not None and torch.max(weights) > self.output_max + eps:
            messages.append("Max weight greater than output_max.")
        if self.output_min is not None and torch.min(weights) < self.output_min - eps:
            messages.append("Min weight less than output_min.")

        diffs = weights[1:] - weights[:-1]
        violation_indices = []

        if self.monotonicity == Monotonicity.INCREASING:
            violation_indices = (diffs < -eps).nonzero().tolist()
        elif self.monotonicity == Monotonicity.DECREASING:
            violation_indices = (diffs > eps).nonzero().tolist()

        violation_indices = [(i[0], i[0] + 1) for i in violation_indices]
        if violation_indices:
            messages.append(f"Monotonicity violated at: {str(violation_indices)}.")

        return messages

    @torch.no_grad()
    def constrain(self) -> None:
        """Jointly projects kernel into desired constraints.

        Uses Dykstra's alternating projection algorithm to jointly project onto all
        given constraints. This algorithm projects with respect to the L2 norm, but it
        approached the norm from the "wrong" side. To ensure that all constraints are
        strictly met, we do final approximate projections that project strictly into the
        feasible space, but this is not an exact projection with respect to the L2 norm.
        Enough iterations make the impact of this approximation negligible.
        """
        constrain_bounds = self.output_min is not None or self.output_max is not None
        constrain_monotonicity = self.monotonicity != Monotonicity.NONE
        num_constraints = sum([constrain_bounds, constrain_monotonicity])

        # We do nothing to the weights in this case
        if num_constraints == 0:
            return

        original_bias, original_heights = self.kernel.data[0:1], self.kernel.data[1:]
        previous_bias_delta = defaultdict(lambda: torch.zeros_like(original_bias))
        previous_heights_delta = defaultdict(lambda: torch.zeros_like(original_heights))

        def apply_bound_constraints(bias, heights):
            previous_bias = bias - previous_bias_delta["BOUNDS"]
            previous_heights = heights - previous_heights_delta["BOUNDS"]
            if constrain_monotonicity:
                bias, heights = self._project_monotonic_bounds(
                    previous_bias, previous_heights
                )
            else:
                bias, heights = self._approximately_project_bounds_only(
                    previous_bias, previous_heights
                )
            previous_bias_delta["BOUNDS"] = bias - previous_bias
            previous_heights_delta["BOUNDS"] = heights - previous_heights
            return bias, heights

        def apply_monotonicity_constraints(heights):
            previous_heights = heights - previous_bias_delta["MONOTONICITY"]
            heights = self._project_monotonicity(previous_heights)
            previous_heights_delta["MONOTONICITY"] = heights - previous_heights
            return heights

        def apply_dykstras_projection(bias, heights):
            if constrain_bounds:
                bias, heights = apply_bound_constraints(bias, heights)
            if constrain_monotonicity:
                heights = apply_monotonicity_constraints(heights)
            return bias, heights

        def finalize_constraints(bias, heights):
            if constrain_monotonicity:
                heights = self._project_monotonicity(heights)
            if constrain_bounds:
                if constrain_monotonicity:
                    bias, heights = self._squeeze_by_scaling(bias, heights)
                else:
                    bias, heights = self._approximately_project_bounds_only(
                        bias, heights
                    )
            return bias, heights

        projected_bias, projected_heights = apply_dykstras_projection(
            original_bias, original_heights
        )
        if num_constraints > 1:
            for _ in range(self.projection_iterations - 1):
                projected_bias, projected_heights = apply_dykstras_projection(
                    projected_bias, projected_heights
                )
            projected_bias, projected_heights = finalize_constraints(
                projected_bias, projected_heights
            )

        self.kernel.data = torch.cat((projected_bias, projected_heights), 0)

    @torch.no_grad()
    def keypoints_inputs(self) -> torch.Tensor:
        """Returns tensor of keypoint inputs."""
        return torch.cat(
            (
                self._interpolation_keypoints,
                self._interpolation_keypoints[-1:] + self._lengths[-1:],
            ),
            0,
        )

    @torch.no_grad()
    def keypoints_outputs(self) -> torch.Tensor:
        """Returns tensor of keypoint outputs."""
        return torch.cumsum(self.kernel.data, 0).T[0]

    ################################################################################
    ############################## PRIVATE METHODS #################################
    ################################################################################

    def _project_monotonic_bounds(
        self, bias: torch.Tensor, heights: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Projects bias and heights into bounds considering monotonicity.

        For computation simplification in the case of decreasing monotonicity, we mirror
        bias and heights and swap-mirror the output bounds. After doing the standard
        projection with resepct to increasing monotonicity, we then mirror everything
        back to get the correct projection.

        Args:
            bias: The bias of the underlying piece-wise linear function.
            heights: The heights of each segment of the underlying piece-wise linear
                function.

        Returns:
            A tuple containing the projected bias and projected heights.
        """
        output_min, output_max = self.output_min, self.output_max
        decreasing = self.monotonicity == Monotonicity.DECREASING
        if decreasing:
            bias, heights = -bias, -heights
            output_min = None if self.output_max is None else -1 * self.output_max
            output_max = None if self.output_min is None else -1 * self.output_min
        if output_max is not None:
            num_heights = heights.size()[0]
            output_max_diffs = output_max - (bias + torch.sum(heights, 0))
            bias_delta = output_max_diffs / (num_heights + 1)
            bias_delta = torch.minimum(bias_delta, torch.tensor(0.0))
            if output_min is not None:
                bias = torch.maximum(bias + bias_delta, torch.tensor(output_min))
                heights_delta = output_max_diffs / num_heights
            else:
                bias += bias_delta
                heights_delta = bias_delta
            heights += torch.minimum(heights_delta, torch.tensor(0.0))
        elif output_min is not None:
            bias = torch.maximum(bias, torch.tensor(output_min))
        if decreasing:
            bias, heights = -bias, -heights
        return bias, heights

    def _approximately_project_bounds_only(
        self, bias: torch.Tensor, heights: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Projects bias and heights without considering monotonicity.

        It is worth noting that this projection is an approximation and is not an exact
        projection with respect to the L2 norm; however, it is sufficiently accurate and
        efficient in practice for non-monotonic functions.

        Args:
            bias: The bias of the underlying piece-wise linear function.
            heights: The heights of each segment of the underlying piece-wise linear
                function.

        Returns:
            A tuple containing the projected bias and projected heights.
        """
        sums = torch.cumsum(torch.cat((bias, heights), 0), 0)
        if self.output_min is not None:
            sums = torch.maximum(sums, torch.tensor(self.output_min))
        if self.output_max is not None:
            sums = torch.minimum(sums, torch.tensor(self.output_max))
        bias = sums[0:1]
        heights = sums[1:] - sums[:-1]
        return bias, heights

    def _project_monotonicity(self, heights: torch.Tensor) -> torch.Tensor:
        """Returns bias and heights projected into desired monotonicity constraints."""
        if self.monotonicity == Monotonicity.INCREASING:
            return torch.maximum(heights, torch.tensor(0.0))
        if self.monotonicity == Monotonicity.DECREASING:
            return torch.minimum(heights, torch.tensor(0.0))
        return heights

    def _squeeze_by_scaling(
        self, bias: torch.Tensor, heights: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Squeezes monotonic calibrators by scaling them into bound constraints.

        It is worth noting that this is not an exact projection with respect to the L2
        norm; however, it maintains convexity, which projection by shift does not.

        Args:
            bias: The bias of the underlying piece-wise linear function.
            heights: The heights of each segment of the underlying piece-wise linear
                function.

        Returns:
            A tuple containing the projected bias and projected heights.
        """
        decreasing = self.monotonicity == Monotonicity.DECREASING
        output_max = self.output_max
        if decreasing:
            if self.output_min is None:
                return bias, heights
            bias, heights = -bias, -heights
            output_max = None if self.output_min is None else -1 * self.output_min
        if output_max is None:
            return bias, heights
        delta = output_max - bias
        scaling_factor = torch.where(
            delta > 0.0001, torch.sum(heights, 0) / delta, torch.ones_like(delta)
        )
        heights /= torch.maximum(scaling_factor, torch.tensor(1.0))
        if decreasing:
            bias, heights = -bias, -heights
        return bias, heights



"""Feature objects for use in models.

To construct a calibrated model, create the calibrated model configuration and pass it
in to the corresponding calibrated model constructor.

Example:
```python
feature_configs = [...]
linear_config = CalibratedLinearConfig(feature_configs, ...)
linear_model = CalibratedLinear(linear_config)
```
"""
import logging
from typing import List, Optional, Tuple, Union

import numpy as np

from .enums import FeatureType, InputKeypointsInit, Monotonicity


# pylint: disable-next=too-few-public-methods
class _Feature:
    """Base configuration for numerical and categorical features.

    Attributes:
        - All `__init__` arguments.
    """

    def __init__(self, feature_name: str, feature_type: FeatureType):
        """Initializes a `_FeatureConfig` instance."""
        self.feature_name = feature_name
        self.feature_type = feature_type


# pylint: disable-next=too-few-public-methods
class NumericalFeature(_Feature):
    """Feature configuration for numerical features.

    Attributes:
        - All `__init__` arguments.
        feature_type: The type of this feature. Always `FeatureType.NUMERICAL`.
        input_keypoints: The input keypoints used for this feature's calibrator. These
            keypoints will be initialized using the given `data` under the desired
            `input_keypoints_init` scheme.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        feature_name: str,
        data: np.ndarray,
        num_keypoints: int = 10,
        input_keypoints_init: InputKeypointsInit = InputKeypointsInit.QUANTILES,
        missing_input_value: Optional[float] = None,
        monotonicity: Monotonicity = Monotonicity.NONE,
        projection_iterations: int = 8,
    ) -> None:
        """Initializes a `NumericalFeatureConfig` instance.

        Args:
            feature_name: The name of the feature. This should match the header for the
                column in the dataset representing this feature.
            data: Numpy array of float-valued data used for calculating keypoint inputs
                and initializing keypoint outputs.
            num_keypoints: The number of keypoints used by the underlying piece-wise
                linear function of a NumericalCalibrator. There will be
                `num_keypoints - 1` total segments.
            input_keypoints_init: The scheme to use for initializing the input
                keypoints. See `InputKeypointsInit` for more details.
            missing_input_value: If provided, this feature's calibrator will learn to
                map all instances of this missing input value to a learned output value.
            monotonicity: Monotonicity constraint for this feature, if any.
            projection_iterations: Number of times to run Dykstra's projection
                algorithm when applying constraints.

        Raises:
            ValueError: If `data` contains NaN values.
            ValueError: If `input_keypoints_init` is invalid.
        """
        super().__init__(feature_name, FeatureType.NUMERICAL)

        if np.isnan(data).any():
            raise ValueError("Data contains NaN values.")

        self.data = data
        self.num_keypoints = num_keypoints
        self.input_keypoints_init = input_keypoints_init
        self.missing_input_value = missing_input_value
        self.monotonicity = monotonicity
        self.projection_iterations = projection_iterations

        sorted_unique_values = np.unique(data)

        if input_keypoints_init == InputKeypointsInit.QUANTILES:
            if sorted_unique_values.size < num_keypoints:
                logging.info(
                    "Observed fewer unique values for feature %s than %d desired "
                    "keypoints. Using the observed %d unique values as keypoints.",
                    feature_name,
                    num_keypoints,
                    sorted_unique_values.size,
                )
                self.input_keypoints = sorted_unique_values
            else:
                quantiles = np.linspace(0.0, 1.0, num=num_keypoints)
                self.input_keypoints = np.quantile(
                    sorted_unique_values, quantiles, method="nearest"
                )
        elif input_keypoints_init == InputKeypointsInit.UNIFORM:
            self.input_keypoints = np.linspace(
                sorted_unique_values[0], sorted_unique_values[-1], num=num_keypoints
            )
        else:
            raise ValueError(f"Unknown input keypoints init: {input_keypoints_init}")


# pylint: disable-next=too-few-public-methods
class CategoricalFeature(_Feature):
    """Feature configuration for categorical features.

    Attributes:
        - All `__init__` arguments.
        feature_type: The type of this feature. Always `FeatureType.CATEGORICAL`.
        category_indices: A dictionary mapping string categories to their index.
        monotonicity_index_pairs: A conversion of `monotonicity_pairs` from string
            categories to category indices. Only available if `monotonicity_pairs` are
            provided.
    """

    def __init__(
        self,
        feature_name: str,
        categories: Union[List[int], List[str]],
        missing_input_value: Optional[float] = None,
        monotonicity_pairs: Optional[List[Tuple[str, str]]] = None,
    ) -> None:
        """Initializes a `CategoricalFeatureConfig` instance.

        Args:
            feature_name: The name of the feature. This should match the header for the
                column in the dataset representing this feature.
            categories: The categories that should be used for this feature. Any
                categories not contained will be considered missing or unknown. If you
                expect to have such missing categories, make sure to
            missing_input_value: If provided, this feature's calibrator will learn to
                map all instances of this missing input value to a learned output value.
            monotonicity_pairs: List of pairs of categories `(category_a, category_b)`
                indicating that the calibrator output for `category_b` should be greater
                than or equal to that of `category_a`.
        """
        super().__init__(feature_name, FeatureType.CATEGORICAL)

        self.categories = categories
        self.missing_input_value = missing_input_value
        self.monotonicity_pairs = monotonicity_pairs

        self.category_indices = {category: i for i, category in enumerate(categories)}
        self.monotonicity_index_pairs = [
            (self.category_indices[a], self.category_indices[b])
            for a, b in monotonicity_pairs or []
        ]



"""Tests for data utilities."""
from datetime import datetime

import numpy as np
import pandas as pd
import pytest
import torch

from sotai import CategoricalFeatureConfig, NumericalFeatureConfig
from sotai.constants import MISSING_CATEGORY_VALUE, MISSING_NUMERICAL_VALUE
from sotai.data import CSVData, determine_feature_types, replace_missing_values
from sotai.enums import FeatureType
from sotai.features import CategoricalFeature, NumericalFeature


@pytest.fixture(name="categorical_data")
def fixture_categorical_data():
    """Returns a mapping for the header and data for categorical data."""
    return {
        "header": "categorical",
        "data": np.array(["a", "b", "c", "d"]),
    }


@pytest.fixture(name="numerical_data")
def fixture_numerical_data():
    """Returns a mapping for the header and data for numerical data."""
    return {
        "header": "numerical",
        "data": np.array([1.0, 2.0, 3.0, 4.0]),
    }


@pytest.fixture(name="target_data")
def fixture_target_data():
    """Returns a mapping for the header and data for target data."""
    return {
        "header": "target",
        "data": np.array([0.2, 0.5, 0.4, 0.9]),
    }


@pytest.fixture(name="unknown_data")
def fixture_unknown_data():
    """Returns a mapping for the header and data for unknown data."""
    return {
        "header": "unknown",
        "data": [datetime(2023, 6, 24) for _ in range(4)],
    }


@pytest.fixture(name="data")
def fixture_data(numerical_data, categorical_data, target_data):
    """Returns a dataframe containing the categorical, numerical, and target data."""
    return pd.DataFrame(
        {
            categorical_data["header"]: categorical_data["data"],
            numerical_data["header"]: numerical_data["data"],
            target_data["header"]: target_data["data"],
        }
    )


@pytest.fixture(name="categories")
def fixture_categories(categorical_data):
    """Returns the categories for the categorical data."""
    return categorical_data["data"][:-1]


@pytest.fixture(name="missing_category")
def fixture_missing_category(categorical_data):
    """Returns a missing category for the categorical data."""
    return categorical_data["data"][-1]


@pytest.fixture(name="missing_input_value")
def fixture_missing_input_value():
    """Returns a missing input value for the categorical data."""
    return -1.0


@pytest.fixture(name="features")
def fixture_features(numerical_data, categorical_data, categories, missing_input_value):
    """Returns a list of features."""
    return [
        NumericalFeature(numerical_data["header"], numerical_data["data"]),
        CategoricalFeature(
            categorical_data["header"],
            categories=categories,
            missing_input_value=missing_input_value,
        ),
    ]


@pytest.fixture(name="feature_configs")
def fixture_feature_configs(numerical_data, categorical_data, categories):
    """Returns a list of feature configs."""
    return {
        "numerical": NumericalFeatureConfig(name=numerical_data["header"]),
        "categorical": CategoricalFeatureConfig(
            name=categorical_data["header"], categories=list(categories)
        ),
    }


def test_replace_missing_values(data, feature_configs):
    """Tests that missing values are properly replaced."""
    data["numerical"].values[-1] = np.nan
    data = replace_missing_values(data, feature_configs)
    assert data["categorical"].values[-1] == MISSING_CATEGORY_VALUE
    assert data["numerical"].values[-1] == MISSING_NUMERICAL_VALUE


def test_determine_feature_types(data, unknown_data):
    """Tests the determination of feature types from data."""
    data[unknown_data["header"]] = unknown_data["data"]
    feature_types = determine_feature_types(data)
    assert feature_types["numerical"] == FeatureType.NUMERICAL
    assert feature_types["categorical"] == FeatureType.CATEGORICAL
    assert feature_types["target"] == FeatureType.NUMERICAL
    assert feature_types["unknown"] == FeatureType.UNKNOWN


@pytest.mark.parametrize("from_filepath", [(True), (False)])
def test_initialization(from_filepath, tmp_path, data):
    """Tests that `CSVData` initialization works as expected."""
    if from_filepath:
        datapath = f"{tmp_path}/data.csv"
        data.to_csv(datapath, index=False)
        csv_data = CSVData(datapath)
        assert csv_data.dataset == datapath
    else:
        csv_data = CSVData(data)
        assert csv_data.dataset.equals(data)

    assert csv_data.data.equals(data)
    assert csv_data.headers == list(data.columns)
    assert csv_data.num_examples == len(data)
    assert csv_data.prepared_data is None
    # pylint: disable=protected-access
    assert csv_data._prepared_data_tensor is None
    assert csv_data._targets_tensor is None
    # pylint: enable=protected-access


def test_call(data, numerical_data, categorical_data, target_data):
    """Tests that the correct column data is returned when called."""
    csv_data = CSVData(data)
    assert np.all(csv_data(numerical_data["header"]) == numerical_data["data"])
    assert np.all(csv_data(categorical_data["header"]) == categorical_data["data"])
    assert np.all(csv_data(target_data["header"]) == target_data["data"])


@pytest.mark.parametrize("include_target", [(True), (False)])
@pytest.mark.parametrize("inplace", [(True), (False)])
def test_prepare(
    include_target,
    inplace,
    data,
    features,
    target_data,
    categorical_data,
    categories,
    missing_category,
    missing_input_value,
):
    """Tests that the data is prepared as expected."""
    csv_data = CSVData(data)
    csv_data.prepare(
        features,
        target_data["header"] if include_target else None,
        inplace=inplace,
    )
    original_categorical_data = pd.Series(csv_data(categorical_data["header"], False))
    categorical_data = pd.Series(csv_data(categorical_data["header"], True))
    if inplace:
        assert (original_categorical_data == categorical_data).all()
    for i, category in enumerate(categories):
        assert (categorical_data[categorical_data == category] == i).all()
    assert (
        categorical_data[categorical_data == missing_category] == missing_input_value
    ).all()
    # pylint: disable=protected-access
    assert csv_data._prepared_data_tensor.size()[1], len(features)
    if include_target:
        assert (
            csv_data._prepared_data_tensor.size()[0]
            == csv_data._targets_tensor.size()[0]
        )
        assert csv_data._targets_tensor.size()[1] == 1
    else:
        assert csv_data._targets_tensor is None
    # pylint: enable=protected-access


@pytest.mark.parametrize(
    "include_target,batch_size,expected_example_batches,expected_target_batches",
    [
        (
            True,
            1,
            [
                torch.tensor([[1, 0]]).double(),
                torch.tensor([[2, 1]]).double(),
                torch.tensor([[3, 2]]).double(),
                torch.tensor([[4, -1]]).double(),
            ],
            [
                torch.tensor([[0.2]]).double(),
                torch.tensor([[0.5]]).double(),
                torch.tensor([[0.4]]).double(),
                torch.tensor([[0.9]]).double(),
            ],
        ),
        (
            True,
            2,
            [
                torch.tensor([[1, 0], [2, 1]]).double(),
                torch.tensor([[3, 2], [4, -1]]).double(),
            ],
            [
                torch.tensor([[0.2], [0.5]]).double(),
                torch.tensor([[0.4], [0.9]]).double(),
            ],
        ),
        (
            True,
            3,
            [
                torch.tensor([[1, 0], [2, 1], [3, 2]]).double(),
                torch.tensor([[4, -1]]).double(),
            ],
            [
                torch.tensor([[0.2], [0.5], [0.4]]).double(),
                torch.tensor([[0.9]]).double(),
            ],
        ),
        (
            False,
            4,
            [
                torch.tensor([[1, 0], [2, 1], [3, 2], [4, -1]]).double(),
            ],
            None,
        ),
    ],
)
def test_batch(
    include_target,
    batch_size,
    expected_example_batches,
    expected_target_batches,
    data,
    features,
    target_data,
):
    """Tests that batches of data are properly generated."""
    csv_data = CSVData(data)
    csv_data.prepare(features, target_data["header"] if include_target else None)
    for i, batch in enumerate(csv_data.batch(batch_size)):
        if include_target:
            example_batch, target_batch = batch
            assert torch.allclose(target_batch, expected_target_batches[i])
        else:
            example_batch = batch
        assert torch.allclose(example_batch, expected_example_batches[i])



"""Tests for api."""
from unittest.mock import patch
import pandas as pd
import numpy as np

from sotai.external import shap


@patch("builtins.open")
@patch("pickle.dump", return_value="test")
@patch("pandas.DataFrame.to_csv", return_value="test")
@patch("sotai.external.post_external_inference", return_value="test_uuid")
def test_shap(mock_post_external_inference, mock_to_csv, mock_dump, mock_open_data):
    """Tests that a pipeline is posted correctly."""

    mock_open_data.return_value.__enter__.return_value = "data"
    test_inference_data = pd.DataFrame([[1, 2, 3], [4, 5, 6]], columns=["a", "b", "c"])

    test_shap_values = np.array([[1, 2, 3], [4, 5, 6]])
    test_base_values = np.array([[1, 2, 3], [4, 5, 6]])

    shap(
        test_inference_data,
        test_shap_values,
        test_base_values,
        "test",
        "target",
        "test",
    )

    mock_post_external_inference.assert_called_with(
        external_shapley_value_name="test",
        shap_filepath="/tmp/sotai/external/shapley_values.csv",
        base_filepath="/tmp/sotai/external/base_values.csv",
        inference_filepath="/tmp/sotai/external/inference_predictions.csv",
        beeswarm_filepath="/tmp/sotai/external/beeswarm_data.pkl",
        scatter_filepath="/tmp/sotai/external/scatter_data.pkl",
        feature_importance_filepath="/tmp/sotai/external/feature_importance_data.pkl",
        target="target",
        dataset_name="test",
    )
    mock_to_csv.assert_called()
    mock_dump.assert_called()
    mock_open_data.assert_called()



"""Layers used in calibrated modeling implemented as `torch.nn.Module`."""
from .categorical_calibrator import CategoricalCalibrator
from .lattice import Lattice
from .linear import Linear
from .numerical_calibrator import NumericalCalibrator



"""SHAP utility functions.

Note that this code is based on code from the SHAP package, so we are including
the license below:

The MIT License (MIT)

Copyright (c) 2018 Scott Lundberg

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import List, Dict, Any

import numpy as np
import pandas as pd
import scipy
from shap import Explanation
from shap.plots._utils import convert_ordering
from shap.utils._exceptions import DimensionError


def calculate_feature_importance(
    shapley_values: np.ndarray, feature_names: List[str]
) -> List[Dict[str, Any]]:
    """Calculates the feature importance from the shapley values.

    Args:
        shapley_values: The shapley values.
        feature_names: The feature names.

    Returns:
        A list of dictionaries containing the feature name and the feature importance value.
    """
    mean_abs_shap_values = np.mean(np.abs(shapley_values), axis=0)
    sorted_indices = np.argsort(mean_abs_shap_values)[::-1]
    sorted_features = [
        {
            "feature": feature_names[i],
            "value": float(mean_abs_shap_values[i]),
        }
        for i in sorted_indices
    ]
    max_display = 10
    if len(sorted_features) > max_display:
        total_sum = 0
        count = 0
        for i in range(max_display - 1, len(sorted_features)):
            total_sum += sorted_features[i]["value"]
            count += 1
        sorted_features[max_display - 1] = {
            "feature": f"Sum of {count} other features",
            "value": total_sum,
        }
    return sorted_features[:max_display]


def calculate_beeswarm(  # pylint: disable-msg=too-many-locals,too-many-branches,too-many-statements
    features: pd.DataFrame, shapley_values: np.ndarray, target: str
) -> List[Dict[str, Any]]:
    """Calculates the beeswarm plot data.

    Args:
        features: The features.
        shapley_values: The shapley values.
        target: The target column.

    Returns:
        A list of dictionaries containing the shapley values, position, color, and name.

        Each dict contains the following fields:

            shaps: The shapley values.
            pos: The position.
            cmap: The color map.
            name: The name.
            vmin: The minimum value.
            vmax: The maximum value.
            c: The color.
    """
    reserved_columns = [target, "logits", "probability"]
    reserved_columns_to_drop = [
        col for col in reserved_columns if col in features.columns
    ]
    if len(reserved_columns_to_drop) > 0:
        features.drop(reserved_columns_to_drop, axis=1, inplace=True)
    feature_names = features.columns.tolist()
    max_display = 10
    color = None
    order = Explanation.abs.mean(0)

    res = []
    sv_shape = shapley_values.shape
    if len(sv_shape) == 1:
        emsg = (
            "The beeswarm plot does not support plotting a single instance, please "
            "pass an explanation matrix with many instances!"
        )
        raise ValueError(emsg)
    if len(sv_shape) > 2:
        emsg = (
            "The beeswarm plot does not support plotting explanations with "
            "instances that have more than one dimension!"
        )
        raise ValueError(emsg)

    # we make a copy here, because later there are places that might modify this array
    values = np.copy(shapley_values)
    if scipy.sparse.issparse(features):
        features = features.toarray()
    # feature_names = shap_exp.feature_names
    # if out_names is None: # TODO: waiting for slicer support
    #     out_names = shap_exp.output_names

    order = convert_ordering(order, values)

    # feature index to category flag
    idx2cat = features.dtypes.astype(str).isin(["object", "category"]).tolist()
    features = features.values

    num_features = values.shape[1]
    if features is not None:
        shape_msg = (
            "The shape of the shap_values matrix does not match the shape "
            "of the provided data matrix."
        )
        if num_features - 1 == features.shape[1]:
            shape_msg += (
                " Perhaps the extra column in the shap_values matrix is the "
                "constant offset? If so, just pass shap_values[:,:-1]."
            )
            raise DimensionError(shape_msg)
        if num_features != features.shape[1]:
            raise DimensionError(shape_msg)

    # determine how many top features we will plot
    if max_display is None:
        max_display = len(feature_names)
    num_features = min(max_display, len(feature_names))

    # iteratively merge nodes until we can cut off the smallest feature values to stay within
    # num_features without breaking a cluster tree
    orig_inds = [[i] for i in range(len(feature_names))]
    feature_order = convert_ordering(order, Explanation(np.abs(values)))
    # here we build our feature names, accounting for the fact that some features
    # might be merged together
    feature_inds = feature_order[:max_display]
    # see how many individual (vs. grouped at the end) features we are plotting
    if num_features < len(values[0]):
        num_cut = np.sum(
            [
                len(orig_inds[feature_order[i]])
                for i in range(num_features - 1, len(values[0]))
            ]
        )
        values[:, feature_order[num_features - 1]] = np.sum(
            [
                values[:, feature_order[i]]
                for i in range(num_features - 1, len(values[0]))
            ],
            0,
        )
        feature_names[
            feature_order[num_features - 1]
        ] = f"Sum of {num_cut} other features"
    row_height = 0.4

    max_rows = 10000
    random_sample = False
    # make the beeswarm dots
    if len(values[:, 0]) > max_rows:
        random_sample = True
        random_indices_for_selection = np.random.choice(
            len(values[:, 0]), max_rows, replace=False
        )
    for pos, i in enumerate(reversed(feature_inds)):
        if random_sample:
            shaps = values[random_indices_for_selection, i]
            fvalues = (
                None if features is None else features[random_indices_for_selection, i]
            )
        else:
            shaps = values[:, i]
            fvalues = None if features is None else features[:, i]
        # Generate 1000 random unique indices
        shap_len = len(shaps)
        inds = np.arange(len(shaps))
        np.random.shuffle(inds)
        if fvalues is not None:
            fvalues = fvalues[inds]
        shaps = shaps[inds]
        colored_feature = True
        try:
            if idx2cat is not None and idx2cat[i]:  # check categorical feature
                colored_feature = False
            else:
                fvalues = np.array(
                    fvalues, dtype=np.float64
                )  # make sure this can be numeric
        except Exception:  # pylint: disable=broad-except
            colored_feature = False
        nbins = 100
        quant = np.round(
            nbins * (shaps - np.min(shaps)) / (np.max(shaps) - np.min(shaps) + 1e-8)
        )
        inds = np.argsort(quant + np.random.randn(shap_len) * 1e-6)
        layer = 0
        last_bin = -1
        y_values = np.zeros(shap_len)
        for ind in inds:
            if quant[ind] != last_bin:
                layer = 0
            y_values[ind] = np.ceil(layer / 2) * ((layer % 2) * 2 - 1)
            layer += 1
            last_bin = quant[ind]
        y_values *= 0.9 * (row_height / np.max(y_values + 1))

        if fvalues is not None and colored_feature:
            # trim the color range, but prevent the color range from collapsing
            vmin = np.nanpercentile(fvalues, 5)
            vmax = np.nanpercentile(fvalues, 95)
            if vmin == vmax:
                vmin = np.nanpercentile(fvalues, 1)
                vmax = np.nanpercentile(fvalues, 99)
                if vmin == vmax:
                    vmin = np.min(fvalues)
                    vmax = np.max(fvalues)
            vmin = min(vmin, vmax)
            assert fvalues.shape[0] == len(
                shaps
            ), "Feature and SHAP matrices must have the same number of rows!"

            # plot the nan fvalues in the interaction feature as grey
            nan_mask = np.isnan(fvalues)

            # plot the non-nan fvalues colored by the trimmed feature value
            cvals = fvalues[np.invert(nan_mask)].astype(np.float64)
            cvals_imp = cvals.copy()
            cvals_imp[np.isnan(cvals)] = (vmin + vmax) / 2.0
            cvals[cvals_imp > vmax] = vmax
            cvals[cvals_imp < vmin] = vmin
            res.append(
                {
                    "shaps": shaps[np.invert(nan_mask)].tolist(),
                    "pos": (pos + y_values[np.invert(nan_mask)]).tolist(),
                    "cmap": color,
                    "vmin": vmin,
                    "vmax": vmax,
                    "c": cvals.tolist(),
                    "name": feature_names[i],
                }
            )
        else:
            res.append(
                {
                    "shaps": shaps.tolist(),
                    "pos": (pos + y_values).tolist(),
                    "cmap": color if colored_feature else "#777777",
                    "name": feature_names[i],
                    "vmin": 0,
                    "vmax": 0,
                }
            )
    return res


def calculate_scatter(
    features: pd.DataFrame, shap_values: np.ndarray
):  # pylint: disable-msg=too-many-locals
    """Calculate scatter plot data for all possible feature combinations.

    Args:
        features: The features.
        shap_values: The shapley values.

    Returns:
        A list of dictionaries containing the scatter plot data for each feature combination.

        Each dict contains the following fields:

            primary_feature_name: The name of the primary feature.
            colorization_feature_name: The name of the colorization feature.
            x_values: The x values.
            y_values: The y values.
            colors: The colors.
            xmin: The minimum x value.
            xmax: The maximum x value.
            ymin: The minimum y value.
            ymax: The maximum y value.
            histogram: The histogram data.
            histogram_bin_edges: The histogram bin edges.
    """
    if len(shap_values.shape) == 1:
        shap_values = shap_values.reshape(1, shap_values.shape[0])

    num_features = shap_values.shape[1]
    scatter_plot_data_list = []

    for primary_index in range(num_features):
        for colorization_index in range(num_features):
            primary_feature_name = features.columns[primary_index]
            colorization_feature_name = features.columns[colorization_index]

            # skip if primary feature is the same as colorization feature
            if primary_feature_name == colorization_feature_name:
                continue

            # Extract primary feature values for x-axis
            oinds = np.arange(shap_values.shape[0])
            np.random.shuffle(oinds)
            x_values = features.iloc[oinds, primary_index].values
            shap_value = shap_values[oinds, primary_index]
            order = np.argsort(-shap_value)

            # Sort the values and handle NaNs
            x_values = x_values[order]
            shap_value = shap_value[order]
            try:
                xv_nan = np.isnan(x_values)
                xv_no_nan = x_values[~xv_nan]

                # Calculate colors based on SHAP values of colorization feature
                cvals = features.iloc[:, colorization_index].astype(np.float64)
                cvals_imp = cvals.copy()
                cvals_imp[np.isnan(cvals)] = np.nanmean(cvals)
            except (TypeError, ValueError):
                continue

            # Plot limits
            xmin = (
                np.nanmin(x_values) - (np.nanmax(x_values) - np.nanmin(x_values)) / 20
            )
            xmax = (
                np.nanmax(x_values) + (np.nanmax(x_values) - np.nanmin(x_values)) / 20
            )
            y1min = np.nanmin(cvals_imp)
            y1max = np.nanmax(cvals_imp)

            # Histogram data
            xvals = np.unique(xv_no_nan)
            if (
                len(xvals) / len(xv_no_nan) < 0.2
                and len(xvals) < 75
                and np.max(xvals) < 75
                and np.min(xvals) >= 0
            ):
                bin_edges = np.arange(np.min(xvals) - 0.5, np.max(xvals) + 1.5) - 0.5
            else:
                if len(xv_no_nan) >= 500:
                    bin_edges = 50
                elif len(xv_no_nan) >= 200:
                    bin_edges = 20
                elif len(xv_no_nan) >= 100:
                    bin_edges = 10
                else:
                    bin_edges = 5
            hist_data, bin_edges_data = np.histogram(
                x_values[~np.isnan(x_values)], bin_edges
            )

            scatter_plot_data_list.append(
                {
                    "primary_feature_name": primary_feature_name,
                    "colorization_feature_name": colorization_feature_name,
                    "x_values": xv_no_nan.tolist(),
                    "y_values": shap_value.tolist(),
                    "colors": cvals_imp.tolist(),
                    "xmin": xmin,
                    "xmax": xmax,
                    "ymin": y1min,
                    "ymax": y1max,
                    "histogram": hist_data.tolist(),
                    "histogram_bin_edges": bin_edges_data.tolist(),
                }
            )

    return scatter_plot_data_list



"""Tests for api."""
from unittest.mock import MagicMock, patch, call
import pandas as pd

from sotai.api import (
    download_prepared_dataset,
    get_dataset_uuids,
    get_inference_results,
    get_inference_status,
    post_inference,
    post_pipeline,
    get_pipeline,
    get_trained_model_metadata,
    post_pipeline_config,
    post_pipeline_feature_configs,
    post_trained_model,
    post_dataset,
    post_external_inference,
    post_hypertune_job,
    post_trained_model_analysis,
)
from sotai.types import HypertuneConfig, LinearConfig
from sotai.enums import LossType
from sotai.constants import SOTAI_API_ENDPOINT, SOTAI_BASE_URL
from .utils import MockResponse


@patch("requests.post", return_value=MockResponse({"uuid": "test_uuid"}))
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_post_pipeline(mock_version, mock_get_api_key, mock_post, fixture_pipeline):
    """Tests that a pipeline is posted correctly.""" ""
    pipeline_response = post_pipeline(fixture_pipeline)

    mock_post.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipelines",
        json={
            "name": "target_classification",
            "target": "target",
            "target_column_type": "classification",
            "primary_metric": "auc",
        },
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )

    mock_get_api_key.assert_called_once()
    mock_version.assert_called()
    mock_version.assert_called_once()
    assert pipeline_response[1] == "test_uuid"
    assert mock_post.call_count == 1


@patch("requests.post", return_value=MockResponse({"uuid": "test_uuid"}))
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_post_pipeline_config(
    mock_version, mock_get_api_key, mock_post, fixture_pipeline_config
):
    """Tests that a pipeline config is posted correctly."""
    pipeline_config_response = post_pipeline_config(
        "test_uuid", fixture_pipeline_config
    )

    mock_post.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipelines/test_uuid/pipeline-configs",
        json={
            "shuffle_data": False,
            "drop_empty_percentage": 80,
            "train_percentage": 60,
            "validation_percentage": 20,
            "test_percentage": 20,
            "pipeline_config_sdk_id": 1,
        },
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()
    assert pipeline_config_response[1] == "test_uuid"


@patch("requests.post", return_value=MockResponse({"uuid": "test_uuid"}))
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_post_feature_configs(
    mock_version,
    mock_get_api_key,
    mock_post,
    fixture_pipeline_config,
    fixture_categories_strs,
    fixture_categories_ints,
):
    """Tests that feature configs are posted correctly."""
    pipeline_config_response = post_pipeline_feature_configs(
        "test_uuid", fixture_pipeline_config.feature_configs
    )

    mock_post.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipeline-configs/test_uuid/feature-configs",
        json=[
            {
                "feature_name": "numerical",
                "feature_type": "numerical",
                "num_keypoints": 10,
                "monotonicity": "increasing",
                "input_keypoints_init": "quantiles",
                "input_keypoints_type": "fixed",
            },
            {
                "feature_name": "categorical_strs",
                "feature_type": "categorical",
                "categories_str": fixture_categories_strs,
            },
            {
                "feature_name": "categorical_ints",
                "feature_type": "categorical",
                "categories_int": fixture_categories_ints,
            },
        ],
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()

    assert pipeline_config_response == "success"


@patch("requests.post", return_value=MockResponse({"uuid": "test_uuid"}))
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_post_trained_model_analysis(
    mock_version, mock_get_api_key, mock_post, fixture_trained_model
):
    """Tests that a trained model is posted correctly."""
    fixture_trained_model.id = 1
    post_trained_model_analysis("test_uuid", fixture_trained_model)

    mock_post.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipeline-configs/test_uuid/analysis",
        json={
            "feature_analyses": [
                {
                    "feature_name": "test",
                    "feature_type": "numerical",
                    "keypoints_inputs_categorical": None,
                    "keypoints_inputs_numerical": [1.0, 2.0, 3.0],
                    "keypoints_outputs": [1.0, 2.0, 3.0],
                    "statistic_max": 2.0,
                    "statistic_mean": 3.0,
                    "statistic_median": 4.0,
                    "statistic_min": 1.0,
                    "statistic_std": 5.0,
                }
            ],
            "model_config": {
                "loss_type": "mse",
                "model_config_name": "Model 1",
                "model_framework": "pytorch",
                "model_type": "linear",
                "primary_metric": "auc",
                "target_column": "target",
                "target_column_type": "classification",
            },
            "overall_model_results": {
                "batch_size": 32,
                "epochs": 100,
                "feature_names": ["test"],
                "learning_rate": 0.001,
                "linear_coefficients": [1.0],
                "runtime_in_seconds": 1.0,
                "test_loss": 1.0,
                "test_primary_metric": 1.0,
                "train_loss_per_epoch": [1.0, 2.0, 3.0],
                "train_primary_metric_per_epoch": [1.0, 2.0, 3.0],
                "validation_loss_per_epoch": [1.0, 2.0, 3.0],
                "validation_primary_metric_per_epoch": [1.0, 2.0, 3.0],
            },
            "trained_model_metadata": {
                "batch_size": 32,
                "epochs": 100,
                "learning_rate": 0.001,
                "test_primary_metric": 1,
                "validation_primary_metric": [3],
                "train_primary_metric": [3],
                "trained_model_sdk_id": 1,
            },
        },
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()


@patch("tarfile.open")
@patch("builtins.open")
@patch("requests.post", return_value=MockResponse({"uuid": "test_uuid"}))
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_post_trained_model(
    mock_version, mock_get_api_key, mock_post, mock_open_data, mock_tarfile_open
):
    """Tests that feature configs are posted correctly."""
    mock_add = MagicMock()
    mock_open_data.return_value.__enter__.return_value = "data"
    mock_tarfile_open.return_value.__enter__.return_value.add = mock_add
    pipeline_response = post_trained_model("/tmp/model", "test_uuid")

    mock_post.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/models",
        files={"file": "data"},
        data={"trained_model_metadata_uuid": "test_uuid"},
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()

    assert pipeline_response == "success"


@patch("builtins.open")
@patch(
    "requests.post",
    return_value=MockResponse({"inference_config_uuid": "test_inference_uuid"}),
)
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_post_inference(
    mock_version,
    mock_get_api_key,
    mock_post,
    mock_open_data,
):
    """Tests that feature configs are posted correctly."""
    mock_open_data.return_value.__enter__.return_value = "data"
    pipeline_response = post_inference("/tmp/model", "test_uuid")

    mock_post.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/inferences",
        files={"file": "data"},
        data={"trained_model_metadata_uuid": "test_uuid"},
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()

    assert pipeline_response[1] == "test_inference_uuid"
    assert pipeline_response[0] == "success"


@patch("requests.get", return_value=MockResponse("initializing", 200))
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_get_inference_status(mock_version, mock_get_api_key, mock_get):
    """Tests that inference config retrieval is handled correctly."""
    inference_status = get_inference_status("test_uuid")

    mock_get.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/inferences/test_uuid/status",
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )

    assert inference_status[1] == "initializing"
    assert inference_status[0] == "success"
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()


@patch("urllib.request.urlretrieve", return_value=None)
@patch("requests.get", return_value=MockResponse("test.com", 200))
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_get_inference_result(
    mock_version, mock_get_api_key, mock_get, mock_urlretrieve
):
    """Tests that inference file retrieval is handled correctly."""
    inference_status = get_inference_results("test_uuid", "/tmp")

    mock_get.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/inferences/test_uuid/download",
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )

    assert inference_status == "success"
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()
    mock_urlretrieve.assert_called_with("test.com", "/tmp/inference_results.csv")


@patch("builtins.open")
@patch(
    "requests.post",
    return_value=MockResponse({"uuid": "test_dataset_uuid"}),
)
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_post_dataset(
    mock_version,
    mock_get_api_key,
    mock_post,
    mock_open_data,
):
    """Tests that feature configs are posted correctly."""
    mock_open_data.return_value = "data"
    pipeline_response = post_dataset(
        "/tmp/dataset/train.csv",
        "/tmp/dataset/test.csv",
        "/tmp/dataset/validation.csv",
        ["age", "chol"],
        [],
        "test_pipeline_uuid",
        1,
    )

    mock_post.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/datasets",
        files=[("files", "data"), ("files", "data"), ("files", "data")],
        data={
            "pipeline_config_uuid": "test_pipeline_uuid",
            "columns": ["age", "chol"],
            "categorical_columns": [],
            "dataset_sdk_id": 1,
        },
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()

    assert pipeline_response[1] == "test_dataset_uuid"
    assert pipeline_response[0] == "success"


@patch(
    "requests.post",
    return_value=MockResponse(
        {"trained_model_metadata_uuids": ["test_uuid_1", "test_uuid_2"]}
    ),
)
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_post_hypertune(
    mock_version, mock_get_api_key, mock_post, fixture_pipeline_config
):
    """Tests that a trained model is posted correctly."""

    hypertune_config = HypertuneConfig(
        epochs=[100],
        batch_sizes=[32],
        learning_rates=[0.001, 0.01],
        loss_type=LossType.BINARY_CROSSENTROPY,
    )
    fixture_pipeline_config.uuid = "test_uuid"
    model_config = LinearConfig()
    post_hypertune_job(
        hypertune_config, fixture_pipeline_config, model_config, "test_dataset_uuid", 1
    )

    mock_post.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipeline-config/test_uuid/hypertune",
        json={
            "dataset_uuid": "test_dataset_uuid",
            "model_config": {
                "model_framework": "pytorch",
                "model_config_name": "Model 1",
                "model_type": "linear",
                "target_column_type": "classification",
                "target_column": "target",
                "primary_metric": "auc",
                "selected_features": [
                    "numerical",
                    "categorical_strs",
                    "categorical_ints",
                ],
                "loss_type": "binary",
                "advanced_options": {
                    "output_min": None,
                    "output_max": None,
                    "output_calibration": False,
                    "output_calibration_num_keypoints": 10,
                    "output_initialization": "quantiles",
                    "output_calibration_input_keypoints_type": "fixed",
                    "use_bias": True,
                },
            },
            "training_config": {
                "epochs_options": [100],
                "batch_size_options": [32],
                "learning_rate_options": [0.001, 0.01],
                "num_parallel_jobs": 4,
                "num_cpus_per_job": 2,
            },
            "next_model_id": 1,
        },
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()


@patch(
    "requests.get",
    return_value=MockResponse(
        {
            "name": "target_classification",
            "target": "target",
            "target_type": "classification",
            "primary_metric": "auc",
            "shuffle_data": True,
            "drop_empty_percentage": 70,
            "dataset_split": {
                "train_percentage": 80,
                "test_percentage": 10,
                "validation_percentage": 10,
            },
            "trained_model_metadata_uuids": ["test_model_uuid"],
            "pipeline_configs": [
                {
                    "uuid": "test_pipeline_config_uuid",
                    "pipeline_config_sdk_id": 1,
                    "shuffle_data": True,
                    "drop_empty_percentage": 70,
                    "train_percentage": 80,
                    "validation_percentage": 10,
                    "test_percentage": 10,
                    "feature_config_list": [
                        {
                            "feature_name": "age",
                            "feature_type": "numerical",
                            "monotonicity": "none",
                            "num_buckets": 0,
                            "num_keypoints": 10,
                            "input_keypoints": "quantiles",
                            "input_keypoints_type": "fixed",
                            "unimodality": "none",
                            "feature_index": 1,
                            "status": "success",
                        },
                        {
                            "feature_name": "sex",
                            "feature_type": "numerical",
                            "monotonicity": "none",
                            "num_buckets": 0,
                            "num_keypoints": 10,
                            "input_keypoints": "quantiles",
                            "input_keypoints_type": "fixed",
                            "unimodality": "none",
                            "feature_index": 2,
                            "status": "success",
                        },
                    ],
                }
            ],
        }
    ),
)
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_get_pipeline(mock_version, mock_get_api_key, mock_get):
    """Tests that a trained model is posted correctly."""

    (
        pipeline_metadata,
        last_pipeline_config,
        pipeline_configs,
        trained_model_uuids,
    ) = get_pipeline("test_uuid")
    mock_get.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipelines/test_uuid",
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()
    assert trained_model_uuids[0] == "test_model_uuid"
    assert pipeline_configs[1].uuid == "test_pipeline_config_uuid"
    assert pipeline_metadata["name"] == "target_classification"
    assert last_pipeline_config == 1


@patch(
    "requests.get",
    return_value=MockResponse(
        {
            "trained_model_metadata": {
                "uuid": "e5433d21-7b48-41e4-93e2-d87c9f33bef8",
                "model_config_id": 1,
                "epochs": 100,
                "learning_rate": 0.01,
                "batch_size": 32,
                "use_sample_weighting": False,
                "train_primary_metric": [0.6956136226654053],
                "validation_primary_metric": [0.7204969525337219],
                "train_loss": [0.5408805320764899],
                "validation_loss": [0.536510161892792],
                "test_primary_metric": 0.7826087474822998,
                "test_loss": 0.4712568339697588,
                "trained_model_sdk_id": None,
            },
            "overall_model_results": {
                "model_config_id": 317,
                "epochs": 100,
                "learning_rate": 0.01,
                "batch_size": 32,
                "test_primary_metric": 0.7826087474822998,
                "test_loss": 0.4712568339697588,
                "train_primary_metric_per_epoch": [
                    0.6437626481056213,
                    0.6544962525367737,
                    0.6522144079208374,
                ],
                "validation_primary_metric_per_epoch": [
                    0.714285671710968,
                    0.7267080545425415,
                    0.7329192757606506,
                ],
                "train_loss_per_epoch": [
                    0.6172197544043537,
                    0.600101311881279,
                    0.5933227655110775,
                ],
                "validation_loss_per_epoch": [
                    0.4814422345862075,
                    0.48776790247917684,
                    0.48770808078051964,
                ],
                "feature_names": [
                    "age",
                    "sex",
                ],
                "feature_importances": [
                    0.21172318895300993,
                    0.3109056738089249,
                    0.11562846497569708,
                    0.0,
                ],
                "linear_coefficients": [
                    0.21172318895300993,
                    0.3109056738089249,
                ],
                "runtime_in_seconds": 1,
            },
            "model_config": {
                "output_min": None,
                "output_max": None,
                "output_calibration": False,
                "output_calibration_num_keypoints": 10,
                "output_initialization": "uniform",
                "output_calibration_input_keypoints_type": "fixed",
                "lattice_size": 2,
                "interpolation": "hypercube",
                "parameterization": "kronecker_factored",
                "num_terms": 2,
                "random_seed": 0,
                "use_bias": True,
                "lattices": "random",
                "model_config_name": "Model 1",
                "model_framework": "pytorch",
                "model_type": "linear",
                "loss_type": "binary",
                "primary_metric": "auc",
                "best_primary_metric": 0.7204969525337219,
                "target_column": "target",
                "target_column_type": "classification",
            },
            "feature_analyses": [
                {
                    "feature_name": "age",
                    "feature_type": "numerical",
                    "statistic_min": 29.0,
                    "statistic_max": 77.0,
                    "statistic_mean": 54.50367647058823,
                    "statistic_median": 55.5,
                    "statistic_std": 8.785850155457481,
                    "keypoints_inputs_numerical": [
                        29.0,
                        39.0,
                        43.0,
                        48.0,
                        52.0,
                        56.0,
                        60.0,
                        65.0,
                        69.0,
                        77.0,
                    ],
                    "keypoints_inputs_categorical": [],
                    "keypoints_outputs": [
                        -2.023148125135339,
                        -1.2120778523877926,
                        -0.6988228320910564,
                        -0.3834010392634607,
                        -0.7148440594259099,
                        -0.2876861882027164,
                        0.06256910105143593,
                        0.6036461743404464,
                        0.2773983506080544,
                        1.5282870069736034,
                    ],
                    "updated_at": "2023-07-24T14:38:21.104913-07:00",
                    "created_at": "2023-07-24T14:38:21.104882-07:00",
                },
                {
                    "feature_name": "sex",
                    "feature_type": "numerical",
                    "statistic_min": 0.0,
                    "statistic_max": 1.0,
                    "statistic_mean": 0.6838235294117647,
                    "statistic_median": 1.0,
                    "statistic_std": 0.4649826986400699,
                    "keypoints_inputs_numerical": [0.0, 1.0],
                    "keypoints_inputs_categorical": [],
                    "keypoints_outputs": [-2.039680783236546, 1.9639668866171704],
                    "updated_at": "2023-07-24T14:38:21.112405-07:00",
                    "created_at": "2023-07-24T14:38:21.112378-07:00",
                },
            ],
            "name": "target_classification",
            "target": "target",
            "target_type": "classification",
            "primary_metric": "auc",
            "shuffle_data": True,
            "drop_empty_percentage": 70,
            "dataset_split": {
                "train_percentage": 80,
                "test_percentage": 10,
                "validation_percentage": 10,
            },
            "trained_model_metadata_uuids": ["test_model_uuid"],
            "pipeline_config": {
                "uuid": "test_pipeline_config_uuid",
                "pipeline_config_sdk_id": 1,
                "shuffle_data": True,
                "drop_empty_percentage": 70,
                "train_percentage": 80,
                "validation_percentage": 10,
                "test_percentage": 10,
                "feature_config_list": [
                    {
                        "feature_name": "age",
                        "feature_type": "numerical",
                        "monotonicity": "none",
                        "num_buckets": 0,
                        "num_keypoints": 10,
                        "input_keypoints": "quantiles",
                        "input_keypoints_type": "fixed",
                        "unimodality": "none",
                        "feature_index": 1,
                        "status": "success",
                    },
                    {
                        "feature_name": "sex",
                        "feature_type": "numerical",
                        "monotonicity": "none",
                        "num_buckets": 0,
                        "num_keypoints": 10,
                        "input_keypoints": "quantiles",
                        "input_keypoints_type": "fixed",
                        "unimodality": "none",
                        "feature_index": 2,
                        "status": "success",
                    },
                ],
            },
        }
    ),
)
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_get_trained_model_metadata(mock_version, mock_get_api_key, mock_get):
    """Tests that a trained model is posted correctly."""

    _, trained_model_metadata = get_trained_model_metadata("test_uuid")
    mock_get.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/trained-model/test_uuid",
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()
    assert (
        trained_model_metadata.training_config.loss_type == LossType.BINARY_CROSSENTROPY
    )
    assert trained_model_metadata.uuid == "test_uuid"


@patch("requests.get", return_value=MockResponse(["test_dataset_uuid"], 200))
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_get_dataset_uuids(mock_version, mock_get_api_key, mock_get):
    """Tests that inference config retrieval is handled correctly."""
    _, dataset_uuids = get_dataset_uuids("test_uuid")

    mock_get.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/pipelines/test_uuid/datasets",
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )

    assert dataset_uuids[0] == "test_dataset_uuid"
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()


@patch("pandas.read_csv", return_value=pd.DataFrame())
@patch("urllib.request.urlretrieve", return_value=None)
@patch(
    "requests.get",
    return_value=MockResponse(
        {
            "train_download_url": "test.com",
            "validation_download_url": "test.com",
            "test_download_url": "test.com",
            "dataset_sdk_id": 1,
            "pipeline_config_sdk_id": 2,
        },
        200,
    ),
)
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_download_prepared_dataset(
    mock_version, mock_get_api_key, mock_get, mock_urlretrieve, mock_pd
):
    """Tests that inference config retrieval is handled correctly."""
    _, dataset = download_prepared_dataset("test_uuid")

    mock_get.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/datasets/test_uuid/download",
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )

    assert dataset.id == 1
    assert dataset.pipeline_config_id == 2

    mock_get_api_key.assert_called_once()
    mock_version.assert_called()
    mock_urlretrieve.assert_has_calls(
        [
            call("test.com", "/tmp/sotai/pipeline/datasets/test_uuid/train.csv"),
            call("test.com", "/tmp/sotai/pipeline/datasets/test_uuid/test.csv"),
            call("test.com", "/tmp/sotai/pipeline/datasets/test_uuid/validation.csv"),
        ]
    )
    mock_pd.assert_has_calls(
        [
            call("/tmp/sotai/pipeline/datasets/test_uuid/train.csv"),
            call("/tmp/sotai/pipeline/datasets/test_uuid/validation.csv"),
            call("/tmp/sotai/pipeline/datasets/test_uuid/test.csv"),
        ]
    )


@patch("builtins.open")
@patch(
    "requests.post",
    return_value=MockResponse({"uuid": "test_shap_uuid"}),
)
@patch("sotai.api.get_api_key", return_value="test_api_key")
@patch("importlib.metadata.version", return_value="0.0.0")
def test_post_external_inference(
    mock_version,
    mock_get_api_key,
    mock_post,
    mock_open_data,
):
    """Tests that feature configs are posted correctly."""
    mock_open_data.return_value = "data"
    pipeline_response = post_external_inference(
        "/tmp/sotai/external/shapley_values.csv",
        "/tmp/sotai/external/base_values.csv",
        "/tmp/sotai/external/inference_predictions.csv",
        "/tmp/sotai/external/beeswarm_data.pkl",
        "/tmp/sotai/external/scatter_data.pkl",
        "/tmp/sotai/external/feature_importance_data.pkl",
        "test",
        "target",
        "dataset_name",
    )

    mock_post.assert_called_with(
        f"{SOTAI_BASE_URL}/{SOTAI_API_ENDPOINT}/shapley-values",
        files=[
            ("files", "data"),
            ("files", "data"),
            ("files", "data"),
            ("files", "data"),
            ("files", "data"),
            ("files", "data"),
        ],
        data={
            "external_shapley_value_name": "test",
            "target": "target",
            "dataset_name": "dataset_name",
        },
        headers={"sotai-api-key": "test_api_key", "sotai-sdk-version": "0.0.0"},
        timeout=10,
    )
    mock_get_api_key.assert_called_once()
    mock_version.assert_called()

    assert pipeline_response[1] == "test_shap_uuid"
    assert pipeline_response[0] == "success"
